import re
from abc import ABC, abstractmethod

import msgspec

from vechord.model import GeminiGenerateRequest
from vechord.provider import GeminiGenerateProvider


class BaseChunker(ABC):
    @abstractmethod
    async def segment(self, text: str) -> list[str]:
        raise NotImplementedError

    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


class RegexChunker(BaseChunker):
    """A simple regex-based chunker."""

    def __init__(
        self,
        size: int = 1536,
        overlap: int = 200,
        separator: str = r"[\n\r\f\v\t?!.;]{1,}",
        concat: str = ". ",
    ):
        self.size = size
        self.overlap = overlap
        self.separator = re.compile(separator)
        self.concatenator = concat

    def name(self) -> str:
        return f"regex_chunk_{self.size}_{self.overlap}"

    def keep_overlap(self, pieces: list[str]) -> list[str]:
        length = 0
        i = len(pieces) - 1
        while i >= 0:
            length += len(pieces[i])
            if length >= self.overlap:
                break
            i -= 1
        return pieces[i + 1 :]

    async def segment(self, text: str) -> list[str]:
        chunks: list[str] = []
        previous: list[str] = []
        current: list[str] = []
        total_length = 0
        pieces = self.separator.split(text)

        for raw_piece in pieces:
            piece = raw_piece.strip()
            if not piece:
                continue
            if total_length + len(piece) > self.size and current:
                chunks.append(self.concatenator.join(previous + current))
                previous = self.keep_overlap(current)
                total_length = sum(len(p) for p in previous)
                current = []
            current.append(piece)
            total_length += len(piece)

            # the new added piece is too long
            if total_length > self.size:
                # pop the overlapping pieces
                overlap_index = 0
                while overlap_index < len(previous):
                    total_length -= len(previous[overlap_index])
                    if total_length <= self.size:
                        break
                    overlap_index += 1
                previous = previous[overlap_index + 1 :]
                chunks.append(self.concatenator.join(previous + current))
                previous = self.keep_overlap(current)
                total_length = sum(len(p) for p in previous)
                current = []

        remaining = self.concatenator.join(previous + current)
        return [*chunks, remaining] if remaining else chunks


class SpacyChunker(BaseChunker):
    """A semantic sentence Chunker based on SpaCy.

    This guarantees the generated chunks are sentences.
    """

    def __init__(self, model: str = "en_core_web_sm"):
        import spacy

        self.model = model
        self.nlp = spacy.load(model, enable=["parser", "tok2vec"])

    def name(self) -> str:
        return f"spacy_chunk_{self.model}"

    async def segment(self, text: str) -> list[str]:
        return [sent.text for sent in self.nlp(text).sents]


class GeminiChunker(BaseChunker, GeminiGenerateProvider):
    """A semantic chunker based on Gemini."""

    def __init__(self, model: str = "gemini-2.5-flash", size: int = 1536):
        super().__init__(model)
        self.prompt = f"""
You are an expert text chunker, skilled at dividing documents into meaningful 
segments while respecting token limits. Your goal is to break down a document into 
chunks that are as semantically coherent as possible, ensuring no chunk exceeds a 
specified token length. Maintain document order. The maximum token length is {size}.
The return format is a list of chunk strings.The document is as follows:
"""
        self.output_token_limit = 65536
        self.regex_chunker = RegexChunker(
            size=self.output_token_limit,
            overlap=0,
            separator=r"[\n\r\f\v\t?!.;]{1,}",
            concat=" ",
        )
        self.json_schema = msgspec.json.schema(list[str])

    def name(self) -> str:
        return f"gemini_chunk_{self.model}"

    async def structure_query(self, prompt: str) -> list[str]:
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_structure_response(
                prompt=prompt,
                schema=self.json_schema,
            )
        )
        chunks = msgspec.json.decode(resp.get_text(), type=list[str])
        return chunks

    async def segment(self, text: str) -> list[str]:
        tokens = len(text)
        if tokens <= self.output_token_limit:
            chunks = await self.structure_query(
                self.prompt + f"\n<document> {text} </document>"
            )
            return chunks

        all_chunks = []
        for chunk in self.regex_chunker.segment(text):
            chunks = await self.structure_query(
                self.prompt + f"\n<document> {chunk} </document>"
            )
            all_chunks.extend(chunks)
        return all_chunks



import contextlib
import math
from contextvars import ContextVar
from typing import Any, Optional, Sequence

import numpy as np
from pgvector.psycopg import register_vector_async
from psycopg import AsyncConnection, sql
from psycopg.errors import DatabaseError
from psycopg_pool import AsyncConnectionPool

from vechord.spec import (
    AnyOf,
    IndexColumn,
    Keyword,
    KeywordIndex,
    MultiVectorIndex,
    UniqueIndex,
    VectorIndex,
)

DEFAULT_TOKENIZER = ("bert_base_uncased", "wiki_tocken")

select_transaction_buffer_conn = ContextVar[Optional[AsyncConnection]](
    "select_transaction_buffer_conn", default=None
)


@contextlib.asynccontextmanager
async def limit_to_transaction_buffer_conn(conn: AsyncConnection):
    """Only the rows inserted in the current transaction are returned."""
    token = select_transaction_buffer_conn.set(conn)
    try:
        yield
    finally:
        select_transaction_buffer_conn.reset(token)


class VechordClient:
    """A PostgreSQL client to access the database.

    Args:
        namespace: used as a prefix for the table name.
        url: the database connection URL.
            e.g. "postgresql://user:password@localhost:5432/dbname"
    """

    def __init__(self, namespace: str, url: str):
        self.ns = namespace
        self.url = url
        self.pool = AsyncConnectionPool(
            self.url, open=False, configure=register_vector_async, max_size=16
        )

    @contextlib.asynccontextmanager
    async def get_connection(self):
        # reuse the connection in a custom transaction buffer
        if conn := select_transaction_buffer_conn.get():
            yield conn
            return

        async with self.pool.connection() as conn:
            yield conn

    async def init_extension(self):
        """Initialize the required PostgreSQL extensions and set the search PATH.

        This should be called once before using the ConnectionPool since the pool
        requires the `vector` type to be created in the Database.
        """
        async with (
            await AsyncConnection.connect(self.url) as conn,
            conn.cursor() as cursor,
        ):
            await cursor.execute("CREATE EXTENSION IF NOT EXISTS vchord CASCADE")
            await cursor.execute("CREATE EXTENSION IF NOT EXISTS vchord_bm25")
            await cursor.execute("CREATE EXTENSION IF NOT EXISTS pg_tokenizer")
            await cursor.execute(
                'SET search_path TO "$user", public, bm25_catalog, tokenizer_catalog'
            )

    async def __aenter__(self):
        await self.init_extension()
        await self.pool.open()
        return self

    async def __aexit__(self, _exc_type, _exc_value, _traceback):
        await self.pool.close()

    async def create_table_if_not_exists(
        self, name: str, schema: Sequence[tuple[str, str]]
    ):
        columns = sql.SQL(", ").join(
            sql.SQL("{col} {typ}").format(
                col=sql.Identifier(col),
                typ=sql.SQL(typ.format(namespace=self.ns)),
            )
            for col, typ in schema
        )
        async with self.get_connection() as conn:
            await conn.execute(
                sql.SQL("CREATE TABLE IF NOT EXISTS {table} ({columns});").format(
                    table=sql.Identifier(f"{self.ns}_{name}"), columns=columns
                )
            )

    async def create_tokenizer(self):
        async with self.get_connection() as conn:
            try:
                for name in DEFAULT_TOKENIZER:
                    await conn.execute(
                        sql.SQL(
                            "SELECT create_tokenizer({name}, $$model={model}$$)"
                        ).format(
                            name=sql.Literal(name),
                            model=sql.Identifier(name),
                        )
                    )
            except DatabaseError as err:
                if "already exists" not in str(err):
                    raise

    async def create_index_if_not_exists(self, name: str, column: IndexColumn):
        if isinstance(column.index, UniqueIndex):
            query = sql.SQL(
                "CREATE UNIQUE INDEX IF NOT EXISTS {index_name} ON {table} "
                "({column}) {config}"
            ).format(
                index_name=sql.Identifier(self._index_name(name, column)),
                table=sql.Identifier(f"{self.ns}_{name}"),
                column=sql.Identifier(column.name),
                config=sql.SQL(column.index.config()),
            )
        else:
            query = sql.SQL(
                "CREATE INDEX IF NOT EXISTS {index_name} ON {table} "
                "USING {index} ({column} {op_name})"
            ).format(
                index_name=sql.Identifier(self._index_name(name, column)),
                table=sql.Identifier(f"{self.ns}_{name}"),
                index=sql.SQL(column.index.index),
                column=sql.Identifier(column.name),
                op_name=sql.SQL(column.index.op_name),
            )
            if config := column.index.config():
                query += sql.SQL(" WITH (options = $${config}$$)").format(
                    config=sql.SQL(config)
                )
        async with self.get_connection() as conn:
            await conn.execute(query)

    def _index_name(self, name: str, column: IndexColumn):
        return f"{self.ns}_{name}_{column.name}_{column.index.name}"

    def _build_conditions(self, kvs: dict[str, Any]):
        if not kvs:
            return sql.SQL("TRUE")
        conditions = []
        for col, val in kvs.items():
            if val is None:
                conditions.append(sql.SQL("{} IS NULL").format(sql.Identifier(col)))
            elif isinstance(val, AnyOf):
                conditions.append(
                    sql.SQL("{} = ANY({})").format(
                        sql.Identifier(col), sql.Literal(val.values)
                    )
                )
            else:
                conditions.append(
                    sql.SQL("{} = {}").format(sql.Identifier(col), sql.Placeholder(col))
                )
        return sql.SQL(" AND ").join(conditions)

    async def select(
        self,
        name: str,
        raw_columns: Sequence[str],
        kvs: Optional[dict[str, Any]] = None,
        from_buffer: bool = False,
        limit: Optional[int] = None,
    ):
        """Select from db table with optional key-value condition or from un-committed
        transaction buffer.

        - `from_buffer`: this ensures the select query only returns the rows that are
            inserted in the current transaction.
        """
        columns = sql.SQL(", ").join(map(sql.Identifier, raw_columns))
        query = sql.SQL("SELECT {columns} FROM {table}").format(
            columns=columns,
            table=sql.Identifier(f"{self.ns}_{name}"),
        )
        if kvs:
            query += sql.SQL(" WHERE {condition}").format(
                condition=self._build_conditions(kvs)
            )
        elif from_buffer:
            query += sql.SQL(" WHERE xmin = pg_current_xact_id()::xid;")
        if limit:
            query += sql.SQL(" LIMIT {}").format(sql.Literal(limit))

        async with self.get_connection() as conn:
            cursor = await conn.execute(query, kvs)
            return await cursor.fetchall()

    @staticmethod
    def _to_placeholder(kv: tuple[str, Any]):
        """Process the `Keyword` type"""
        key, value = kv
        if isinstance(value, Keyword):
            return sql.SQL("tokenize({}, {})").format(
                sql.Placeholder(key), sql.Literal(value._model)
            )
        return sql.Placeholder(key)

    async def insert(self, name: str, values: dict):
        columns = sql.SQL(", ").join(map(sql.Identifier, values))
        placeholders = sql.SQL(", ").join(map(self._to_placeholder, values.items()))
        query = sql.SQL(
            "INSERT INTO {table} ({columns}) VALUES ({placeholders});"
        ).format(
            table=sql.Identifier(f"{self.ns}_{name}"),
            columns=columns,
            placeholders=placeholders,
        )
        async with self.get_connection() as conn:
            await conn.execute(query, values)

    async def copy_bulk(self, name: str, values: Sequence[dict], types: Sequence[str]):
        columns = sql.SQL(", ").join(map(sql.Identifier, values[0]))
        query = sql.SQL(
            "COPY {table} ({columns}) FROM STDIN WITH (FORMAT BINARY)"
        ).format(
            table=sql.Identifier(f"{self.ns}_{name}"),
            columns=columns,
        )
        async with self.get_connection() as conn:
            cursor = conn.cursor()
            async with cursor.copy(query) as copy:
                copy.set_types(types=types)
                for value in values:
                    await copy.write_row(tuple(value.values()))

    async def delete(self, name: str, kvs: dict):
        async with self.get_connection() as conn:
            if kvs:
                await conn.execute(
                    sql.SQL("DELETE FROM {table} WHERE {condition};").format(
                        table=sql.Identifier(f"{self.ns}_{name}"),
                        condition=self._build_conditions(kvs),
                    ),
                    kvs,
                )
            else:
                await conn.execute(
                    sql.SQL("DELETE FROM {table};").format(
                        table=sql.Identifier(f"{self.ns}_{name}")
                    )
                )

    async def query_vec(  # noqa: PLR0913
        self,
        name: str,
        vec_col: IndexColumn[VectorIndex],
        vec: np.ndarray,
        return_fields: Sequence[str],
        topk: int = 10,
        probe: Optional[int] = None,
    ):
        columns = sql.SQL(", ").join(map(sql.Identifier, return_fields))
        if (
            probe is None
            and vec_col.index.lists is not None
            and vec_col.index.lists > 1
        ):
            probe = math.ceil(vec_col.index.lists / 16)
        set_probe = sql.SQL("SET LOCAL vchordrq.probes = {};").format(
            sql.Literal(probe or "")
        )
        async with self.get_connection() as conn:
            cursor = conn.cursor()
            await cursor.execute(set_probe)
            await cursor.execute(
                sql.SQL(
                    "SELECT {columns} FROM {table} ORDER BY {vec_col} {op} %s LIMIT %s;"
                ).format(
                    table=sql.Identifier(f"{self.ns}_{name}"),
                    columns=columns,
                    op=sql.SQL(vec_col.index.op_symbol),
                    vec_col=sql.Identifier(vec_col.name),
                ),
                (vec, topk),
            )
            return await cursor.fetchall()

    async def query_multivec(  # noqa: PLR0913
        self,
        name: str,
        multivec_col: IndexColumn[MultiVectorIndex],
        vec: np.ndarray,
        maxsim_refine: int,
        probe: Optional[int],
        return_fields: Sequence[str],
        topk: int = 10,
    ):
        columns = sql.SQL(", ").join(map(sql.Identifier, return_fields))
        if (
            probe is None
            and multivec_col.index.lists is not None
            and multivec_col.index.lists > 1
        ):
            probe = math.ceil(multivec_col.index.lists / 16)
        set_probe = sql.SQL("SET LOCAL vchordrq.probes = {};").format(
            sql.Literal(probe or "")
        )
        set_refine = sql.SQL("SET LOCAL vchordrq.maxsim_refine = {};").format(
            sql.Literal(maxsim_refine)
        )
        async with self.get_connection() as conn:
            cursor = conn.cursor()
            await cursor.execute(set_probe)
            await cursor.execute(set_refine)
            await cursor.execute(
                sql.SQL(
                    "SELECT {columns} FROM {table} ORDER BY {multivec_col} @# %s LIMIT %s;"
                ).format(
                    table=sql.Identifier(f"{self.ns}_{name}"),
                    columns=columns,
                    multivec_col=sql.Identifier(multivec_col.name),
                ),
                (vec, topk),
            )
            return await cursor.fetchall()

    async def query_keyword(  # noqa: PLR0913
        self,
        name: str,
        keyword_col: IndexColumn[KeywordIndex],
        keyword: str,
        return_fields: Sequence[str],
        tokenizer: str,
        topk: int = 10,
    ):
        columns = sql.SQL(", ").join(map(sql.Identifier, return_fields))
        async with self.get_connection() as conn:
            cursor = await conn.execute(
                sql.SQL(
                    "SELECT {columns} FROM {table} ORDER BY {keyword_col} <&> "
                    "to_bm25query({index}, tokenize(%s, {tokenizer})) LIMIT %s;"
                ).format(
                    table=sql.Identifier(f"{self.ns}_{name}"),
                    columns=columns,
                    index=sql.Literal(self._index_name(name, keyword_col)),
                    tokenizer=sql.Literal(tokenizer),
                    keyword_col=sql.Identifier(keyword_col.name),
                ),
                (keyword, topk),
            )
            return await cursor.fetchall()

    async def drop(self, name: str):
        async with self.get_connection() as conn:
            await conn.execute(
                sql.SQL("DROP TABLE IF EXISTS {table} CASCADE;").format(
                    table=sql.Identifier(f"{self.ns}_{name}")
                )
            )



from os import environ

import numpy as np
import pytest

from vechord.client import VechordClient
from vechord.log import logger
from vechord.registry import VechordRegistry
from vechord.spec import Vector


@pytest.fixture
def anyio_backend():
    # trio is not supported by psycopg
    # https://github.com/psycopg/psycopg/issues/29
    return "asyncio"


URL = "127.0.0.1"
# for local container development environment, use the host machine's IP
if environ.get("REMOTE_CONTAINERS", "") == "true" or environ.get("USER", "") == "envd":
    URL = "172.17.0.1"
TEST_POSTGRES = f"postgresql://postgres:postgres@{URL}:5432/"


@pytest.fixture(name="registry")
async def fixture_registry(request):
    namespace = request.node.obj.__name__
    tables = request.param or ()
    async with VechordRegistry(namespace, TEST_POSTGRES, tables=tables) as registry:
        yield registry

    # when the registry is used as Falcon middleware, the lifespan protocol will close
    # the connection, so we need to re-create the database connection to cleanup
    logger.debug("clearing storage...")
    async with VechordClient(namespace, TEST_POSTGRES) as client:
        for table in tables:
            await client.drop(table.name())


DenseVector = Vector[128]


def gen_vector(dim: int = 128) -> DenseVector:
    rng = np.random.default_rng()
    return DenseVector(rng.random((dim,), dtype=np.float32))



import asyncio
from typing import Any

import httpx

from vechord.typing import Self

# https://ai.google.dev/gemini-api/docs/rate-limits#tier-1
GEMINI_GENERATE_RPS = 16.66
GEMINI_EMBEDDING_RPS = 0.6
# https://docs.voyageai.com/docs/rate-limits
VOYAGE_EMBEDDING_RPS = 33.33
# https://jina.ai/api-dashboard/rate-limit
JINA_EMBEDDING_RPS = 8.33


class RateLimitTransport(httpx.AsyncHTTPTransport):
    def __init__(self, max_per_second: float = 5, **kwargs) -> None:
        """
        Async HTTP transport with rate limit.

        Args:
            max_per_second: Maximum number of requests per second.

        Other args are passed to httpx.AsyncHTTPTransport.
        """
        self.interval = 1 / max_per_second
        self.next_start_time = 0
        super().__init__(**kwargs)

    async def notify_task_start(self):
        """
        https://github.com/florimondmanca/aiometer/blob/358976e0b60bce29b9fe8c59807fafbad3e62cbc/src/aiometer/_impl/meters.py#L57
        """
        loop = asyncio.get_running_loop()
        while True:
            now = loop.time()
            next_start_time = max(self.next_start_time, now)
            until_now = next_start_time - now
            if until_now <= self.interval:
                break
            await asyncio.sleep(max(0, until_now - self.interval))
        self.next_start_time = max(self.next_start_time, now) + self.interval

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        await self.notify_task_start()
        return await super().handle_async_request(request)

    async def __aenter__(self) -> Self:
        await self.notify_task_start()
        return await super().__aenter__()

    async def __aexit__(self, *args: Any) -> None:
        await super().__aexit__(*args)



import itertools
from collections.abc import Callable, Iterable
from contextlib import contextmanager
from os import environ
from typing import TYPE_CHECKING, Annotated, Any, Optional
from uuid import UUID

import msgspec

from vechord.chunk import BaseChunker, GeminiChunker, RegexChunker
from vechord.client import VechordClient, limit_to_transaction_buffer_conn
from vechord.embedding import (
    BaseMultiModalEmbedding,
    BaseTextEmbedding,
    GeminiDenseEmbedding,
    JinaDenseEmbedding,
    JinaMultiModalEmbedding,
    OpenAIDenseEmbedding,
    VoyageDenseEmbedding,
    VoyageMultiModalEmbedding,
)
from vechord.entity import GeminiEntityRecognizer
from vechord.errors import RequestError
from vechord.extract import GeminiExtractor, LlamaParseExtractor
from vechord.model import (
    GraphEntity,
    GraphRelation,
    InputType,
    ResourceRequest,
    RunAck,
    RunRequest,
)
from vechord.rerank import CohereReranker
from vechord.spec import (
    AnyOf,
    DefaultDocument,
    Keyword,
    KeywordIndex,
    PrimaryKeyUUID,
    Table,
    Vector,
    VectorIndex,
    _DefaultChunk,
)
from vechord.typing import Self

if TYPE_CHECKING:
    from vechord.registry import VechordRegistry


class GraphIndex(msgspec.Struct):
    """Graph index for entities and relations extracted from the text/image."""


class IndexOption:
    def __init__(self, vector=None, keyword=None, graph=None):
        self.vector = (
            msgspec.convert(vector, VectorIndex) if vector is not None else None
        )
        self.keyword = (
            msgspec.convert(keyword, KeywordIndex) if keyword is not None else None
        )
        self.graph = msgspec.convert(graph, GraphIndex) if graph is not None else None


class VectorSearchOption(msgspec.Struct, kw_only=True):
    topk: int
    probe: Optional[int] = None


class KeywordSearchOption(msgspec.Struct, kw_only=True):
    topk: int


class GraphSearchOption(msgspec.Struct, kw_only=True):
    topk: int


class SearchOption:
    def __init__(self, vector=None, keyword=None, graph=None):
        self.vector = msgspec.convert(vector, VectorSearchOption) if vector else None
        self.keyword = (
            msgspec.convert(keyword, KeywordSearchOption) if keyword else None
        )
        self.graph = msgspec.convert(graph, GraphSearchOption) if graph else None


class _Entity(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    chunk_uuids: list[UUID]
    text: str
    label: str
    description: str = ""
    vec: Vector[1]  # as a placeholder, will be replaced by the actual vector type


class _Relation(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    source: UUID
    target: UUID
    description: str
    vec: Vector[1]  # as a placeholder, will be replaced by the actual vector type


PROVIDER_MAP: dict[str, dict[str, Any]] = {
    "chunk": {
        "regex": RegexChunker,
        "gemini": GeminiChunker,
    },
    "text-emb": {
        "gemini": GeminiDenseEmbedding,
        "jina": JinaDenseEmbedding,
        "openai": OpenAIDenseEmbedding,
        "voyage": VoyageDenseEmbedding,
    },
    "multimodal-emb": {
        "voyage": VoyageMultiModalEmbedding,
        "jina": JinaMultiModalEmbedding,
    },
    "ocr": {"gemini": GeminiExtractor, "llamaparse": LlamaParseExtractor},
    "rerank": {"cohere": CohereReranker},
    "graph": {"gemini": GeminiEntityRecognizer},
    "index": {"vectorchord": IndexOption},
    "search": {"vectorchord": SearchOption},
}


@contextmanager
def set_api_key_in_env(keys: Iterable[str], values: Iterable[Optional[str]]):
    old_values = {key: environ.get(key) for key in keys if key}
    for key, value in zip(keys, values, strict=False):
        if not key:
            continue
        if value is None:
            environ.pop(key)
        environ[key] = value
    try:
        yield
    finally:
        for key, old_value in old_values.items():
            if old_value is None:
                environ.pop(key, None)
            else:
                environ[key] = old_value


def find_uid_by_text(text: str, ents: list[_Entity]) -> Optional[UUID]:
    for ent in ents:
        if ent.text == text:
            return ent.uid
    return None


class DynamicPipeline(msgspec.Struct, kw_only=True):
    chunk: Optional[BaseChunker] = None
    text_emb: Optional[BaseTextEmbedding] = None
    multimodal_emb: Optional[BaseMultiModalEmbedding] = None
    ocr: Optional[GeminiExtractor] = None
    rerank: Optional[CohereReranker] = None
    index: Optional[IndexOption] = None
    search: Optional[SearchOption] = None
    graph: Optional[GeminiEntityRecognizer] = None

    def __post_init__(self):
        if not (self.text_emb or self.multimodal_emb):
            raise RequestError("No embedding provider specified in the request")
        if self.index is None and self.search is None:
            raise RequestError("No `index` or `search` option specified in the request")
        if self.index and self.index.graph and not self.graph:
            raise RequestError("Graph index requires a graph provider")
        if self.index and self.index.vector is None:
            raise RequestError("Vector index is required if `index` is specified")
        if self.search and not (self.text_emb or self.multimodal_emb):
            raise RequestError("Search requires at least one embedding provider")

    @classmethod
    def from_steps(cls, steps: list[ResourceRequest]) -> Self:
        calls = {}
        for step in steps:
            if step.kind not in PROVIDER_MAP:
                raise RequestError(f"Unsupported provider kind: {step.kind}")

            provider = PROVIDER_MAP[step.kind].get(step.provider)
            if not provider:
                raise RequestError(
                    f"Unsupported provider: {step.provider} for kind: {step.kind}"
                )

            api_name, api_key = "", None
            args = step.args
            if "api_key" in args:
                api_name = f"{step.provider.upper()}_API_KEY"
                api_key = args.pop("api_key", None)
            with set_api_key_in_env((api_name,), (api_key,)):
                calls[(step.kind).replace("-", "_")] = provider(**args)
        return msgspec.convert(calls, DynamicPipeline)

    async def run(self, request: RunRequest, vr: "VechordRegistry"):
        """Run the dynamic pipeline with the given request."""
        if self.index:
            return await self.run_index(request, vr)
        elif self.search:
            return await self.run_search(request, vr)
        else:
            raise RequestError("No valid pipeline configuration found")

    @staticmethod
    def _convert_from_extracted_graph(
        uid: UUID,
        ents: list[GraphEntity],
        rels: list[GraphRelation],
        ent_cls: type[Table],
        rel_cls: type[Table],
    ):
        converted_ents = [
            ent_cls(
                chunk_uuids=[uid],
                text=ent.text,
                label=ent.label,
                description=ent.description,
                vec=None,  # will be set later
            )
            for ent in ents
        ]
        converted_rels = [
            rel_cls(
                source=find_uid_by_text(rel.source.text, converted_ents),
                target=find_uid_by_text(rel.target.text, converted_ents),
                description=f"{rel.source.text} {rel.description} {rel.target.text}",
                vec=None,  # will be set later
            )
            for rel in rels
        ]
        return converted_ents, converted_rels

    async def run_index(self, request: RunRequest, vr: "VechordRegistry") -> RunAck:  # noqa: PLR0912
        dim = (
            self.text_emb.get_dim() if self.text_emb else self.multimodal_emb.get_dim()
        )
        enable_keyword_index = self.index.keyword is not None
        enable_graph_index = self.index.graph is not None
        vec_index_type = Annotated[Vector[dim], self.index.vector]

        Chunk = msgspec.defstruct(
            "Chunk", (("vec", vec_index_type),), bases=(_DefaultChunk,)
        )
        # use the default vector index for Entity and Relation
        Entity = msgspec.defstruct("Entity", (("vec", Vector[dim]),), bases=(_Entity,))
        Relation = msgspec.defstruct(
            "Relation", (("vec", Vector[dim]),), bases=(_Relation,)
        )
        tables = [DefaultDocument, Chunk]
        if enable_graph_index:
            tables.extend([Entity, Relation])

        await vr.init_table_index(tables=tables)

        # run the pipeline
        sentences, chunks, ents, rels = [], [], [], []
        doc = DefaultDocument(text="")
        if self.multimodal_emb:
            chunks.append(
                Chunk(
                    doc_id=doc.uid,
                    text="",
                    keyword=None,
                    vec=await self.multimodal_emb.vectorize_multimodal_chunk(
                        request.data
                    ),
                )
            )
        else:
            if request.input_type is InputType.TEXT:
                doc.text = request.data.decode("utf-8")
            elif self.ocr:
                if request.input_type is InputType.PDF:
                    doc.text = await self.ocr.extract_pdf(request.data)
                elif request.input_type is InputType.IMAGE:
                    doc.text = await self.ocr.extract_image(request.data)
            elif self.graph:
                img_ents, img_rels = await self.graph.recognize_image(request.data)
                conv_ents, conv_rels = self._convert_from_extracted_graph(
                    doc.uid, img_ents, img_rels, Entity, Relation
                )
                ents.extend(conv_ents)
                rels.extend(conv_rels)
            else:
                raise RequestError(
                    f"No OCR provider for input type: {request.input_type}"
                )

            if self.chunk:
                sentences.extend(await self.chunk.segment(doc.text))
            elif doc.text:
                sentences.append(doc.text)
            for sent in sentences:
                chunk = Chunk(
                    vec=await self.text_emb.vectorize_chunk(sent),
                    doc_id=doc.uid,
                    text=sent,
                    keyword=None if not enable_keyword_index else Keyword(sent),
                )
                if self.graph and request.input_type is InputType.TEXT:
                    chunk_ents, chunk_rels = await self.graph.recognize_with_relations(
                        sent
                    )
                    conv_ents, conv_rels = self._convert_from_extracted_graph(
                        chunk.uid, chunk_ents, chunk_rels, Entity, Relation
                    )
                    ents.extend(conv_ents)
                    rels.extend(conv_rels)
                chunks.append(chunk)

        async with (
            vr.client.get_connection() as conn,
            limit_to_transaction_buffer_conn(conn),
        ):
            await vr.insert(doc)
            for chunk in chunks:
                await vr.insert(chunk)
            if self.index.graph:
                await self.graph_insert(
                    ents=ents, rels=rels, ent_cls=Entity, rel_cls=Relation, vr=vr
                )
            return RunAck(name=request.name, msg="succeed", uid=doc.uid)

    async def graph_insert(
        self,
        ents: list[_Entity],
        rels: list[_Relation],
        ent_cls: type[Table],
        rel_cls: type[Table],
        vr: "VechordRegistry",
    ):
        """Insert entities and relations into the graph index."""
        ent_map: dict[str, _Entity] = {}
        for ent in ents:
            if ent.text not in ent_map:
                ent_map[ent.text] = ent
            else:
                ent_map[ent.text].chunk_uuids.extend(ent.chunk_uuids)
                ent_map[ent.text].description += f"\n{ent.description}"
        for ent in ent_map.values():
            exist_ent = await vr.select_by(
                ent_cls.partial_init(text=ent.text),
                fields=("uid", "description", "chunk_uuids"),
            )
            if exist_ent:
                exist = exist_ent[0]
                ent.chunk_uuids.extend(exist.chunk_uuids)
                ent.description += f"\n{exist.description}"
                await vr.remove_by(ent_cls.partial_init(uid=exist.uid))
            ent.vec = await self.text_emb.vectorize_chunk(
                f"{ent.text}\n{ent.description}"
            )
            await vr.insert(ent)

        relation_map: dict[str, _Relation] = {}
        for rel in rels:
            key = "|".join(sorted(map(str, [rel.source, rel.target])))
            if key not in relation_map:
                relation_map[key] = rel
            else:
                relation_map[key].description += f"\n{rel.description}"
        for rel in relation_map.values():
            exist_rel = await vr.select_by(
                rel_cls.partial_init(source=rel.source, target=rel.target),
                fields=("uid", "description"),
            )
            if exist_rel:
                exist = exist_rel[0]
                rel.description += f"\n{exist.description}"
                await vr.remove_by(rel_cls.partial_init(uid=exist.uid))
            rel.vec = await self.text_emb.vectorize_chunk(f"{rel.description}")
            await vr.insert(rel)

    async def run_search(self, request: RunRequest, vr: "VechordRegistry"):
        query = request.data.decode("utf-8")

        # for type hint and compatibility
        class Chunk(_DefaultChunk):
            pass

        class Entity(_Entity):
            pass

        class Relation(_Relation):
            pass

        retrieved: list[Chunk] = []
        if self.search.vector:
            vec = (
                await self.text_emb.vectorize_query(query)
                if self.text_emb
                else await self.multimodal_emb.vectorize_multimodal_query(text=query)
            )
            retrieved.extend(
                await vr.search_by_vector(
                    Chunk, vec, self.search.vector.topk, probe=self.search.vector.probe
                )
            )
        if self.search.keyword:
            retrieved.extend(
                await vr.search_by_keyword(Chunk, query, self.search.keyword.topk)
            )
        if self.search.graph:
            retrieved.extend(
                await self.graph_search(query, Chunk, Entity, Relation, vr)
            )
        if self.rerank:
            indices = await self.rerank.rerank(
                query, [chunk.text for chunk in retrieved]
            )
            retrieved = [retrieved[i] for i in indices]
        return retrieved

    async def graph_search(
        self,
        query: str,
        chunk_cls: type[Table],
        ent_cls: type[Table],
        rel_cls: type[Table],
        vr: "VechordRegistry",
    ):
        ents, rels = await self.graph.recognize_with_relations(query)
        if rels:
            rel_text = " ".join(rel.description for rel in rels)
            similar_rels = await vr.search_by_vector(
                rel_cls,
                await self.text_emb.vectorize_query(rel_text),
                topk=self.search.graph.topk,
            )
            ent_uuids = deduplicate_uid(
                itertools.chain.from_iterable(
                    (rel.source, rel.target) for rel in similar_rels
                )
            )
            ents.extend(
                await vr.select_by(
                    ent_cls.partial_init(uid=AnyOf(ent_uuids)),
                    fields=("text", "description"),
                )
            )
        if not ents:
            return []
        ent_text = " ".join(f"{ent.text} {ent.description}" for ent in ents)
        similar_ents = await vr.search_by_vector(
            ent_cls,
            await self.text_emb.vectorize_query(ent_text),
            topk=self.search.graph.topk,
        )
        chunk_uuids = deduplicate_uid(
            itertools.chain.from_iterable(ent.chunk_uuids for ent in similar_ents),
        )
        chunks = await vr.select_by(
            chunk_cls.partial_init(uid=AnyOf(chunk_uuids)),
            fields=("text", "doc_id", "uid"),
        )
        return chunks


def deduplicate_uid(uuids: Iterable[UUID], limit: Optional[int] = None) -> list[UUID]:
    """Maintain the order of the occurrence of UUIDs and deduplicate them."""
    uuids = {uid: None for uid in uuids}
    return list(uuids.keys())[:limit]


class VechordPipeline:
    """Set up the pipeline to run multiple functions in a transaction.

    Args:
        client: :class:`VectorChordClient` to be used for the transaction.
        steps: a list of functions to be run in the pipeline. The first function
            will be used to accept the input, and the last function will be used
            to return the output. The rest of the functions will be used to
            process the data in between. The functions will be run in the order
            they are defined in the list.
    """

    def __init__(self, client: VechordClient, steps: list[Callable]):
        self.client = client
        self.steps = steps

    async def run(self, *args, **kwargs) -> Any:
        """Execute the pipeline in a transactional manner.

        All the `args` and `kwargs` will be passed to the first function in the
        pipeline. The pipeline will run in *one* transaction, and all the `inject`
        can only see the data inserted in this transaction (to guarantee only the
        new inserted data will be processed in this pipeline).

        This will also return the final result of the last function in the pipeline.
        """
        async with (
            self.client.get_connection() as conn,
            limit_to_transaction_buffer_conn(conn),
        ):
            # only the 1st one can accept input (could be empty)
            await self.steps[0](*args, **kwargs)
            for func in self.steps[1:-1]:
                await func()
            return await self.steps[-1]()



import hashlib
import os
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path

from vechord.log import logger
from vechord.model import Document


class BaseLoader(ABC):
    @abstractmethod
    def load(self) -> list[Document]:
        raise NotImplementedError

    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


class LocalLoader(BaseLoader):
    """Load documents from local file system."""

    def __init__(self, path: str, include: list[str] | None = None):
        self.path = Path(path)
        self.include = set(ext.lower() for ext in include or [".txt"])

    def name(self) -> str:
        return "local"

    def load(self) -> list[Document]:
        res = []
        for root, _dirs, files in os.walk(self.path):
            for file in files:
                filepath = Path(root) / file
                ext = filepath.suffix.lower()
                if ext not in self.include:
                    logger.debug("exclude file %s", file)
                    continue
                data = filepath.read_bytes()
                res.append(
                    Document(
                        path=str(filepath),
                        data=data,
                        ext=ext,
                        digest=hashlib.sha256(data).hexdigest(),
                        updated_at=datetime.fromtimestamp(filepath.stat().st_mtime),
                        source=self.name(),
                    )
                )
        return res


class S3Loader(BaseLoader):
    def __init__(self, bucket: str, prefix: str, include: list[str] | None = None):
        self.bucket = bucket
        self.prefix = prefix
        self.include = set(ext.lower() for ext in include or [".txt"])

    def name(self) -> str:
        return f"s3_{self.bucket}_{self.prefix}"

    def load(self) -> list[Document]:
        # TODO: implement S3 loader
        raise NotImplementedError



from vechord.errors import (
    APIKeyUnsetError,
    DecodeStructuredOutputError,
    HTTPCallError,
    RequestError,
    TimeoutError,
    UnexpectedResponseError,
    extract_safe_err_msg,
)


def test_extracted_msg():
    # other errors should return empty string
    assert extract_safe_err_msg(ValueError("what happened")) == ""

    # hide the real error message
    for exc in [
        DecodeStructuredOutputError,
        UnexpectedResponseError,
        HTTPCallError,
    ]:
        assert extract_safe_err_msg(exc("some error")) == str(exc("doesn't matter"))

    assert "UNIVERSAL" in extract_safe_err_msg(APIKeyUnsetError("UNIVERSAL"))
    assert "leaked" in extract_safe_err_msg(RequestError("leaked"))
    assert "timeout" in extract_safe_err_msg(TimeoutError("timeout"))



from enum import Enum
from uuid import UUID

import msgspec

from vechord.typing import Self


# https://docs.cloud.llamaindex.ai/llamaparse/features/supported_document_types
# Currently only
class LlamaCloudMimeType(str, Enum):
    # Documents
    PDF = "application/pdf"
    EPUB = "application/epub+zip"
    RTF = "application/rtf"
    XML = "application/xml"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    TXT = "text/plain"
    HTML = "text/html"
    MARKDOWN = "text/markdown"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    CSV = "text/csv"
    JSON = "application/json"

    # Images
    JPG = "image/jpeg"
    JPEG = "image/jpeg"
    PNG = "image/png"
    GIF = "image/gif"
    BMP = "image/bmp"
    SVG = "image/svg+xml"
    TIFF = "image/tiff"
    WEBP = "image/webp"

    # Audio/Video
    MP3 = "audio/mpeg"
    MP4 = "video/mp4"
    MPEG = "video/mpeg"
    MPGA = "audio/mpeg"
    M4A = "audio/mp4"
    WAV = "audio/wav"
    WEBM = "video/webm"


class LlamaCloudJobStatus(str, Enum):
    PENDING = "PENDING"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"
    PARTIAL_SUCCESS = "PARTIAL_SUCCESS"
    CANCELLED = "CANCELLED"


class LlamaCloudParseRequest(msgspec.Struct, kw_only=True):
    filename: str
    content: bytes
    mime_type: LlamaCloudMimeType = LlamaCloudMimeType.PDF

    @classmethod
    def from_image(cls, img: bytes, mime_type: LlamaCloudMimeType) -> Self:
        # FIXME: Add real filename
        return cls(filename="image.jpg", content=img, mime_type=mime_type)

    @classmethod
    def from_pdf(cls, pdf: bytes, mime_type: LlamaCloudMimeType) -> Self:
        return cls(filename="document.pdf", content=pdf, mime_type=mime_type)


class LlamaCloudParseResponse(msgspec.Struct, kw_only=True):
    id: UUID
    status: LlamaCloudJobStatus
    error_code: int | None = None
    error_message: str | None = None



from enum import Enum
from typing import Any, Literal
from uuid import UUID

import msgspec


class ResourceRequest(msgspec.Struct, kw_only=True):
    kind: Literal[
        "ocr", "chunk", "text-emb", "multimodal-emb", "rerank", "index", "search"
    ]
    provider: str
    args: dict[str, Any] = msgspec.field(default_factory=dict)


class InputType(str, Enum):
    TEXT = "text"
    PDF = "pdf"
    IMAGE = "image/jpeg"


class RunRequest(msgspec.Struct, kw_only=True, frozen=True):
    """Request to run a dynamic pipeline.

    Possible `data` types:
        - text
        - PDF
        - image/jpeg
    """

    name: str
    data: bytes
    input_type: InputType = InputType.TEXT
    steps: list[ResourceRequest] = msgspec.field(default_factory=list)


class RunAck(msgspec.Struct, kw_only=True, frozen=True):
    name: str
    msg: str
    uid: UUID



import os
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Literal, Optional

import httpx
import msgspec
import numpy as np

from vechord.log import logger
from vechord.model import (
    GeminiEmbeddingRequest,
    JinaEmbeddingRequest,
    SparseEmbedding,
    VoyageEmbeddingRequest,
    VoyageMultiModalEmbeddingRequest,
)
from vechord.provider import (
    GeminiEmbeddingProvider,
    JinaEmbeddingProvider,
    VoyageEmbeddingProvider,
)


class VecType(Enum):
    DENSE = auto()
    SPARSE = auto()
    KEYWORD = auto()


class BaseEmbedding(ABC):
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_dim(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def vec_type(self) -> VecType:
        raise NotImplementedError


class BaseTextEmbedding(BaseEmbedding):
    @abstractmethod
    async def vectorize_chunk(self, text: str) -> np.ndarray:
        raise NotImplementedError

    async def vectorize_query(self, text: str) -> np.ndarray:
        return await self.vectorize_chunk(text)


class BaseMultiModalEmbedding(BaseEmbedding):
    @abstractmethod
    async def vectorize_multimodal_chunk(
        self, text: str, image: Optional[bytes] = None, image_url: Optional[str] = None
    ) -> np.ndarray:
        raise NotImplementedError

    async def vectorize_multimodal_query(
        self, text: str, image: Optional[bytes] = None, image_url: Optional[str] = None
    ) -> np.ndarray:
        return await self.vectorize_multimodal_chunk(text, image, image_url)


class SpacyDenseEmbedding(BaseTextEmbedding):
    """Spacy Dense Embedding.

    The default small model is unlikely to be useful for real applications.
    It's recommended to use the small model for development and testing
    purpose only.
    """

    def __init__(self, model: str = "en_core_web_sm", dim: int = 96):
        import spacy

        self.nlp = spacy.load(model, enable=["tok2vec"])
        self.dim = dim
        self.model = model

    def name(self) -> str:
        return f"spacy_emb_{self.model}"

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self):
        return VecType.DENSE

    async def vectorize_chunk(self, text: str) -> np.ndarray:
        doc = self.nlp(text)
        return doc.vector


class GeminiDenseEmbedding(BaseTextEmbedding, GeminiEmbeddingProvider):
    """Gemini Dense Embedding. (limit to **8,192** tokens)

    Args:
        model: Gemini embedding model name
        dim: embedding dimension, could be [768, 1536, 3072]
    """

    def __init__(
        self,
        model: str = "gemini-embedding-exp-03-07",
        dim: Literal[768, 1536, 3072] = 3072,
    ):
        super().__init__(model, dim)

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.DENSE

    def name(self) -> str:
        return f"gemini_emb_{self.model}_{self.dim}"

    async def vectorize_chunk(self, text: str) -> np.ndarray:
        resp = await self.query(
            GeminiEmbeddingRequest.from_text_with_type(text, "RETRIEVAL_DOCUMENT")
        )
        return resp.get_emb()

    async def vectorize_query(self, text):
        resp = await self.query(
            GeminiEmbeddingRequest.from_text_with_type(text, "RETRIEVAL_QUERY")
        )
        return resp.get_emb()


class JinaDenseEmbedding(BaseTextEmbedding, JinaEmbeddingProvider):
    """Jina Dense Embedding. (limit to **32,768** tokens for v4, **8,192** for v3)

    Args:
        model: Jina embedding model name, could be "jina-embeddings-v4" or "jina-embeddings-v3"
        dim: embedding dimension, up to 2048
    """

    def __init__(self, model: str = "jina-embeddings-v4", dim: int = 2048):
        super().__init__(model, dim)

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.DENSE

    def name(self) -> str:
        return f"jina_emb_{self.model}_{self.dim}"

    async def vectorize_chunk(self, text: str) -> np.ndarray:
        resp = await self.query(
            JinaEmbeddingRequest.from_text(text, "retrieval.passage", self.model)
        )
        return resp.get_emb()

    async def vectorize_query(self, text: str) -> np.ndarray:
        resp = await self.query(
            JinaEmbeddingRequest.from_text(text, "retrieval.query", self.model)
        )
        return resp.get_emb()


class JinaMultiModalEmbedding(BaseMultiModalEmbedding, JinaEmbeddingProvider):
    """Jina MultiModal Dense Embedding, limit to **32,768** tokens for v4.

    Args:
        model: Jina embedding model name, could be "jina-embeddings-v4"
        dim: embedding dimension, up to 2048
    """

    def __init__(self, model="jina-embeddings-v4", dim=2048):
        super().__init__(model, dim)

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.DENSE

    def name(self) -> str:
        return f"jina_emb_{self.model}_{self.dim}"

    async def vectorize_multimodal_chunk(
        self, text: str, image: Optional[bytes] = None, image_url: Optional[str] = None
    ) -> np.ndarray:
        req = await self.query(
            JinaEmbeddingRequest.from_text_image(
                text=text,
                image=image,
                image_url=image_url,
                task="retrieval.passage",
                model=self.model,
            )
        )
        return req.get_emb()

    async def vectorize_multimodal_query(
        self, text: str, image: Optional[bytes] = None, image_url: Optional[str] = None
    ) -> np.ndarray:
        req = await self.query(
            JinaEmbeddingRequest.from_text_image(
                text=text,
                image=image,
                image_url=image_url,
                task="retrieval.query",
                model=self.model,
            )
        )
        return req.get_emb()


class VoyageDenseEmbedding(BaseTextEmbedding, VoyageEmbeddingProvider):
    def __init__(
        self, model: str = "voyage-3.5", dim: Literal[256, 512, 1024, 2048] = 1024
    ):
        super().__init__(model, dim)

    def name(self):
        return f"voyage_emb_{self.model}_{self.dim}"

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.DENSE

    async def vectorize_chunk(self, text):
        resp = await self.query(
            VoyageEmbeddingRequest.from_text(
                text=text, input_type="document", model=self.model, dim=self.dim
            )
        )
        return resp.get_emb()

    async def vectorize_query(self, text):
        resp = await self.query(
            VoyageEmbeddingRequest.from_text(
                text=text, input_type="query", model=self.model, dim=self.dim
            )
        )
        return resp.get_emb()


class VoyageMultiModalEmbedding(BaseMultiModalEmbedding, VoyageEmbeddingProvider):
    """Voyage Multimodal Embedding.

    Accepts text, image (as bytes), or image_url.

    Limits:
        - image: less than **16 million pixels** or **20 MB** in size
    """

    def __init__(self, model="voyage-multimodal-3", dim=1024):
        super().__init__(model, dim)
        self.url = "https://api.voyageai.com/v1/multimodalembeddings"

    def name(self):
        return f"voyage_multimodal_emb_{self.model}_{self.dim}"

    def get_dim(self):
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.DENSE

    async def vectorize_multimodal_chunk(
        self,
        image: Optional[bytes] = None,
        text: Optional[str] = None,
        image_url: Optional[str] = None,
    ):
        resp = await self.query(
            VoyageMultiModalEmbeddingRequest.build(
                text=text,
                image=image,
                image_url=image_url,
                input_type="document",
                model=self.model,
            )
        )
        return resp.get_emb()

    async def vectorize_multimodal_query(
        self,
        image: Optional[bytes] = None,
        text: Optional[str] = None,
        image_url: Optional[str] = None,
    ):
        resp = await self.query(
            VoyageMultiModalEmbeddingRequest.build(
                text=text,
                image=image,
                image_url=image_url,
                input_type="query",
                model=self.model,
            )
        )
        return resp.get_emb()


class OpenAIDenseEmbedding(BaseTextEmbedding):
    """OpenAI Dense Embedding."""

    def __init__(self, model: str = "text-embedding-3-large", dim: int = 3072):
        key = os.environ.get("OPENAI_API_KEY")
        if not key:
            raise ValueError("env OPENAI_API_KEY not set")

        from openai import AsyncOpenAI

        self.client = AsyncOpenAI()
        self.model = model
        self.dim = dim

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.DENSE

    def name(self) -> str:
        return f"openai_emb_{self.model}_{self.dim}"

    async def vectorize_chunk(self, text: str) -> np.ndarray:
        resp = await self.client.embeddings.create(
            model=self.model, input=text, dimensions=self.dim
        )
        return np.array(resp.data[0].embedding, dtype=np.float32)


class SpladePPSparseEmbedding:
    """Splade++ Sparse Embedding."""

    def __init__(self, url: str, dim: int = 30522, timeout_sec: int = 10):
        self.dim = dim
        self.client = httpx.Client(base_url=url, timeout=timeout_sec)
        self.decoder = msgspec.json.Decoder(type=[SparseEmbedding])

    def get_dim(self) -> int:
        return self.dim

    def vec_type(self) -> VecType:
        return VecType.SPARSE

    def name(self) -> str:
        return f"spladepp_emb_{self.dim}"

    def vectorize_chunk(self, text: str | list[str]) -> list[SparseEmbedding]:
        resp = self.client.post("/inference", json=text)
        if resp.is_error:
            logger.info(
                "failed to call Splade++ sparse emb: [%d] %s",
                resp.status_code,
                resp.content,
            )
            resp.raise_for_status()

        sparse = self.decoder.decode(resp.content)
        return sparse



from typing import Any, Optional, TypeVar

import falcon
import msgspec
import numpy as np
from defspec import OpenAPI, RenderTemplate
from falcon.asgi import App, Request, Response

from vechord.errors import extract_safe_err_msg
from vechord.log import logger
from vechord.model import RunAck, RunRequest
from vechord.pipeline import DynamicPipeline
from vechord.registry import Table, VechordPipeline, VechordRegistry
from vechord.spec import _DefaultChunk

T = TypeVar("T")
M = TypeVar("M", bound=msgspec.Struct)


def vechord_schema_hook(cls: type):
    if method := getattr(cls, "__json_schema__", None):
        return method()
    raise NotImplementedError()


def vechord_encode_hook(obj: Any) -> Any:
    if method := getattr(obj, "__json_encode__", None):
        return method(obj)
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    raise NotImplementedError(
        f"Object {obj} does not implement __json_encode__ method."
    )


def vechord_decode_hook(obj_type: type[T], data: Any) -> T:
    if method := getattr(obj_type, "__json_decode__", None):
        return method(data)
    raise NotImplementedError(
        f"Type {obj_type} does not implement __json_decode__ method."
    )


async def validate_request(spec: type[M], req: Request, resp: Response) -> Optional[M]:
    decoder = (
        msgspec.msgpack if req.content_type == falcon.MEDIA_MSGPACK else msgspec.json
    )
    try:
        if req.method == "GET":
            request = msgspec.convert(req.params, spec)
        else:
            request = decoder.decode(
                await req.stream.read(), type=spec, dec_hook=vechord_decode_hook
            )
    except (msgspec.ValidationError, msgspec.DecodeError) as err:
        logger.info(
            "failed to decode the request '%s' body %s: %s", req.path, spec, err
        )
        resp.status = falcon.HTTP_422
        resp.text = f"Validation error: {err}"
        resp.content_type = falcon.MEDIA_TEXT
        return None
    return request


async def uncaught_exception_handler(
    req: Request, resp: Response, exc: Exception, params: dict
):
    logger.warning(
        "exception from endpoint '%s'",
        req.path,
        exc_info=exc,
    )
    description = extract_safe_err_msg(exc)
    raise falcon.HTTPError(falcon.HTTP_500, title=req.path, description=description)


class HealthCheck:
    async def on_get(self, req: Request, resp: Response):
        resp.status = falcon.HTTP_200
        resp.content_type = falcon.MEDIA_TEXT
        resp.text = "Ok"


class TableResource:
    def __init__(self, table: type[Table], registry: VechordRegistry):
        self.table_cls = table
        self.registry = registry

    async def on_get(self, req: Request, resp: Response):
        table = self.table_cls.partial_init(**req.params)
        rows = await self.registry.select_by(obj=table)
        resp.data = msgspec.json.encode(rows, enc_hook=vechord_encode_hook)

    async def on_post(self, req: Request, resp: Response):
        table = await validate_request(self.table_cls, req, resp)
        if table is None:
            return

        await self.registry.insert(table)
        resp.status = falcon.HTTP_CREATED

    async def on_delete(self, req: Request, resp: Response):
        table = self.table_cls.partial_init(**req.params)
        await self.registry.remove_by(obj=table)


class PipelineResource:
    def __init__(self, pipeline: VechordPipeline):
        self.pipeline = pipeline
        self.decoder = msgspec.json.Decoder(dec_hook=vechord_decode_hook)

    async def on_post(self, req: Request, resp: Response):
        json = self.decoder.decode(await req.stream.read())
        if not isinstance(json, dict):
            raise falcon.HTTPBadRequest(
                title="Invalid request",
                description="Request must be a JSON Dict",
            )
        await self.pipeline.run(**json)


class RunResource:
    def __init__(self, registry: VechordRegistry):
        self.registry = registry

    async def on_post(self, req: Request, resp: Response):
        request = await validate_request(RunRequest, req, resp)
        if request is None:
            return
        pipe = DynamicPipeline.from_steps(request.steps)
        res = await pipe.run(request, self.registry)
        if res:
            encoder = (
                msgspec.msgpack
                if req.content_type == falcon.MEDIA_MSGPACK
                else msgspec.json
            )
            resp.data = encoder.encode(res, enc_hook=vechord_encode_hook)


class OpenAPIResource:
    def __init__(
        self, tables: list[type[Table]], include_pipeline: bool = True
    ) -> None:
        self.openapi = OpenAPI()
        self.openapi.register_route("/", "get", summary="health check")
        self.openapi.register_route(
            "/api/run",
            "post",
            summary="run the pipeline",
            request_type=RunRequest,
            response_type=RunAck | list[_DefaultChunk],
            schema_hook=vechord_schema_hook,
        )
        if include_pipeline:
            self.openapi.register_route(
                "/api/pipeline", "post", summary="run the pipeline"
            )
        for table in tables:
            path = f"/api/table/{table.name()}"
            self.openapi.register_route(
                path,
                "get",
                "get the table with partial attributes",
                query_type=table,
                schema_hook=vechord_schema_hook,
            )
            self.openapi.register_route(
                path,
                "delete",
                "delete table records according to partial attributes",
                query_type=table,
                schema_hook=vechord_schema_hook,
            )
            self.openapi.register_route(
                path,
                "post",
                "insert a new record to the table",
                request_type=table,
                request_content_type="json",
                schema_hook=vechord_schema_hook,
            )
        self.spec = self.openapi.to_json()

    async def on_get(self, req: Request, resp: Response):
        resp.content_type = falcon.MEDIA_JSON
        resp.data = self.spec


class OpenAPIRender:
    def __init__(self, spec_url: str, template: RenderTemplate) -> None:
        self.template = template.value.format(spec_url=spec_url)

    async def on_get(self, req: Request, resp: Response):
        resp.content_type = falcon.MEDIA_HTML
        resp.text = self.template


def create_web_app(
    registry: VechordRegistry, pipeline: Optional[VechordPipeline] = None
) -> App:
    """Create a `Falcon` ASGI application for the given registry.

    This includes the:

    - health check [GET](/)
    - tables [GET/POST/DELETE](/api/table/{table_name})
    - OpenAPI spec and Swagger UI [GET](/openapi/swagger)
    - optional: pipeline in a transaction [POST](/api/pipeline)
    """
    app = App(middleware=registry)
    app.req_options.strip_url_path_trailing_slash = True
    app.add_route("/", HealthCheck())
    for table in registry.tables:
        app.add_route(
            f"/api/table/{table.name()}",
            TableResource(table=table, registry=registry),
        )
    if pipeline is not None:
        app.add_route("/api/pipeline", PipelineResource(pipeline=pipeline))
    app.add_route("/api/run", RunResource(registry=registry))
    app.add_route(
        "/openapi/spec.json", OpenAPIResource(registry.tables, pipeline is not None)
    )
    app.add_route(
        "/openapi/swagger", OpenAPIRender("/openapi/spec.json", RenderTemplate.SWAGGER)
    )
    app.add_error_handler(Exception, uncaught_exception_handler)
    return app



from abc import ABC, abstractmethod

from vechord.model import GeminiGenerateRequest
from vechord.provider import GeminiGenerateProvider


class BaseAugmenter(ABC):
    @abstractmethod
    def reset(self, doc: str):
        """Cache the document for augmentation."""
        raise NotImplementedError

    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    async def augment_context(self, doc: str, chunks: list[str]) -> list[str]:
        raise NotImplementedError

    @abstractmethod
    async def augment_query(self, doc: str, chunks: list[str]) -> list[str]:
        raise NotImplementedError

    @abstractmethod
    async def summarize_doc(self, doc: str) -> str:
        raise NotImplementedError


class GeminiAugmenter(BaseAugmenter, GeminiGenerateProvider):
    """Gemini Augmenter.

    Context caching is only available for stable models with fixed versions.
    Minimal cache token is 32768.
    """

    def __init__(self, model: str = "gemini-2.5-flash"):
        super().__init__(model)

    def name(self) -> str:
        return f"gemini_augment_{self.model}"

    async def augment(self, doc: str, chunks: list[str], prompt: str) -> list[str]:
        res = []
        for chunk in chunks:
            context = prompt.format(chunk=chunk)
            context = f"<document>\n{doc}\n</document>\n" + context
            resp = await self.query(GeminiGenerateRequest.from_prompt(context))
            res.append(resp.get_text().strip())
        return res

    async def augment_context(self, doc: str, chunks: list[str]) -> list[str]:
        """Generate the contextual chunks.

        Args:
            doc: document text
            chunks: list of chunks to augment
        """
        prompt = (
            "Here is the chunk we want to situate within the whole document \n"
            "<chunk>\n{chunk}\n</chunk>\n"
            "Please give a short succinct context to situate this chunk within "
            "the overall document for the purposes of improving search retrieval "
            "of the chunk. Answer only with the succinct context and nothing else."
        )
        return await self.augment(doc, chunks, prompt)

    async def augment_query(self, doc: str, chunks: list[str]) -> list[str]:
        """Generate the queries for chunks.

        Args:
            doc: document text
            chunks: list of chunks to augment
        """
        prompt = (
            "Here is the chunk we want to ask questions about \n"
            "<chunk>\n{chunk}\n</chunk>\n"
            "Please ask questions about this chunk based on the overall document "
            "for the purposes of improving search retrieval of the chunk. "
            "Answer only with the question and nothing else."
        )
        return await self.augment(doc, chunks, prompt)

    async def summarize_doc(self, doc: str) -> str:
        """Summarize the document."""
        prompt = (
            "Summarize the provided document concisely while preserving its key "
            "ideas, main arguments, and essential details. Ensure clarity and "
            "coherence, avoiding unnecessary repetition."
            f"\n<document>{doc}</document>\n"
        )
        resp = await self.query(GeminiGenerateRequest.from_prompt(prompt))
        return resp.get_text().strip()



from argparse import ArgumentParser

import uvicorn

from vechord.registry import VechordRegistry
from vechord.service import create_web_app


def build_parser() -> ArgumentParser:
    parser = ArgumentParser(description="Run vechord as a dynamic pipeline service.")
    parser.add_argument(
        "--db",
        type=str,
        default="postgresql://postgres:postgres@localhost:5432/",
        help="VectorChord DB URL.",
    )
    parser.add_argument("--host", type=str, default="localhost", help="Service host.")
    parser.add_argument("--port", type=int, default=8000, help="Service port.")
    return parser


def main():
    parser = build_parser()
    args = parser.parse_args()

    vr = VechordRegistry("cli", args.db)
    app = create_web_app(vr)
    uvicorn.run(app, host=args.host, port=args.port)



import base64
import re
from typing import Literal, Optional

import msgspec
import numpy as np

from vechord.errors import RequestError, UnexpectedResponseError
from vechord.typing import Self


def pascal_to_snake(s: str) -> str:
    s = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", s)
    s = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s)
    return s.lower()


class VoyageEmbedding(msgspec.Struct, kw_only=True):
    embedding: list[float] | bytes  # msgspec will handle the base64 encoding/decoding
    index: int


class VoyageEmbeddingResponse(msgspec.Struct, kw_only=True):
    data: list[VoyageEmbedding]

    def get_emb(self) -> np.ndarray:
        """Get the first embedding as a numpy array."""
        if not self.data or not self.data[0].embedding:
            raise UnexpectedResponseError("empty embedding data")
        emb = self.data[0].embedding
        if isinstance(emb, list):
            return np.array(emb, dtype=np.float32)
        return np.frombuffer(emb, dtype=np.float32)


VOYAGE_INPUT_TYPE = Literal["query", "document"]


class VoyageEmbeddingRequest(msgspec.Struct, kw_only=True):
    model: str
    input_text: str | list[str] = msgspec.field(name="input")
    input_type: VOYAGE_INPUT_TYPE = "document"
    truncation: bool = True
    output_dimension: int
    output_dtype: Literal["float", "int8", "uint8", "binary", "ubinary"] = "float"
    encoding_format: Optional[Literal["base64"]] = "base64"

    @classmethod
    def from_text(
        cls, text: str, input_type: VOYAGE_INPUT_TYPE, model: str, dim: int
    ) -> Self:
        return VoyageEmbeddingRequest(
            model=model,
            input_text=text,
            input_type=input_type,
            output_dimension=dim,
        )


class Text(msgspec.Struct, tag=pascal_to_snake):
    text: str


class ImageBase64(msgspec.Struct, tag=pascal_to_snake):
    image_base64: str


class ImageURL(msgspec.Struct, tag=pascal_to_snake):
    image_url: str


class MultiModalInput(msgspec.Struct, tag=pascal_to_snake):
    content: list[Text | ImageBase64 | ImageURL]


class VoyageMultiModalEmbeddingRequest(msgspec.Struct, kw_only=True):
    model: str
    inputs: list[MultiModalInput]
    input_type: VOYAGE_INPUT_TYPE = "document"
    truncation: bool = True
    encoding_format: Optional[Literal["base64"]] = "base64"

    @classmethod
    def build(
        cls,
        text: Optional[str],
        image_url: Optional[str],
        image: Optional[bytes],
        model: str,
        input_type: VOYAGE_INPUT_TYPE,
    ) -> Self:
        if not (text or image_url or image):
            raise RequestError(
                "At least one of text, image_url, or image must be provided."
            )
        contents = []
        if text:
            contents.append(Text(text=text))
        if image_url:
            contents.append(ImageURL(image_url=image_url))
        if image:
            contents.append(
                ImageBase64(
                    image_base64=f"data:image/jpeg;base64,{base64.b64encode(image).decode('utf-8')}",
                )
            )
        return VoyageMultiModalEmbeddingRequest(
            model=model,
            inputs=[MultiModalInput(content=contents)],
            input_type=input_type,
        )



import asyncio
from collections.abc import AsyncGenerator
from datetime import datetime, timezone
from functools import partial
from typing import Annotated, Optional

import msgspec
import pytest
from psycopg.errors import UniqueViolation
from psycopg.types.json import Jsonb

from tests.conftest import DenseVector, gen_vector
from vechord.spec import (
    AnyOf,
    ForeignKey,
    Keyword,
    MultiVectorIndex,
    PrimaryKeyAutoIncrease,
    PrimaryKeyUUID,
    Table,
    UniqueIndex,
    VectorIndex,
)

pytestmark = pytest.mark.anyio


class Document(Table, kw_only=True):
    uid: PrimaryKeyAutoIncrease | None = None
    title: str = ""
    text: str
    updated_at: datetime = msgspec.field(
        default_factory=partial(datetime.now, timezone.utc)
    )


class Chunk(Table, kw_only=True):
    uid: PrimaryKeyAutoIncrease | None = None
    doc_id: Annotated[int, ForeignKey[Document.uid]]
    text: str
    vector: DenseVector
    keyword: Keyword


class AnnotatedChunk(Table, kw_only=True):
    uid: Optional[PrimaryKeyAutoIncrease] = None
    text: str
    vector: Annotated[DenseVector, VectorIndex(distance="cos", lists=3)]


class Sentence(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    text: str
    vector: list[DenseVector]


class UniqueTable(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    sid: Annotated[str, UniqueIndex()]


class SubTable(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    text: str
    foreign_key: Annotated[str, ForeignKey[UniqueTable.sid]]


class JsonTable(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    text: str
    data: Jsonb


class Image(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    text: str
    vecs: Annotated[list[DenseVector], MultiVectorIndex(lists=2)]
    dataset: Optional[str] = None
    oid: Optional[int] = None


Tockenizer = Keyword.with_model("wiki_tocken")


class OtherTokenizer(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    text: str
    keyword: Tockenizer


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Document,)], indirect=True)
async def test_insert_select_remove(registry):
    docs = [Document(text="hello world"), Document(text="hello there")]
    for doc in docs:
        await registry.insert(doc)

    # select all
    inserted = await registry.select_by(Document.partial_init(), fields=["text"])
    assert len(inserted) == len(docs)
    assert inserted[0].text == "hello world"
    assert inserted[1].text == "hello there"

    # select with limit
    one = await registry.select_by(Document.partial_init(), limit=1)
    assert len(one) == 1

    # select by id
    first = await registry.select_by(Document.partial_init(uid=1))
    assert len(first) == 1
    assert first[0].text == "hello world"

    # select by AnyOf
    await registry.insert(Document(text="hello again"))
    targets = [1, 3]
    multi = await registry.select_by(Document.partial_init(uid=AnyOf(targets)))
    assert len(multi) == len(targets)
    assert multi[0].text == "hello world"
    assert multi[1].text == "hello again"
    await registry.remove_by(Document.partial_init(uid=AnyOf(targets)))

    # remove by id
    await registry.remove_by(Document.partial_init(uid=2))
    assert len(await registry.select_by(Document.partial_init())) == 0


@pytest.mark.db
@pytest.mark.parametrize("registry", [(AnnotatedChunk,)], indirect=True)
async def test_annotated_index(registry):
    num = 100
    topk = 5
    for text in (f"hello {i}" for i in range(num)):
        await registry.insert(AnnotatedChunk(text=text, vector=gen_vector()))

    inserted = await registry.select_by(AnnotatedChunk.partial_init(), fields=["text"])
    assert len(inserted) == num

    res = await registry.search_by_vector(AnnotatedChunk, gen_vector(), topk=topk)
    assert len(res) == topk


@pytest.mark.db
@pytest.mark.parametrize("registry", [(UniqueTable, SubTable)], indirect=True)
async def test_unique_index(registry):
    await registry.insert(UniqueTable(sid="id_0"))
    with pytest.raises(UniqueViolation):
        await registry.insert(UniqueTable(sid="id_0"))
    await registry.insert(SubTable(text="hello", foreign_key="id_0"))
    await registry.remove_by(UniqueTable.partial_init(sid="id_0"))
    assert len(await registry.select_by(UniqueTable.partial_init())) == 0
    assert len(await registry.select_by(SubTable.partial_init())) == 0


@pytest.mark.db
@pytest.mark.parametrize("registry", [(OtherTokenizer,)], indirect=True)
async def test_keyword_tokenizer(registry):
    num = 20
    topk = 5
    for text in (f"hello {i}" for i in range(num)):
        await registry.insert(OtherTokenizer(text=text, keyword=Tockenizer(text)))

    inserted = await registry.select_by(OtherTokenizer.partial_init(), fields=["text"])
    assert len(inserted) == num

    res = await registry.search_by_keyword(OtherTokenizer, "hello", topk=topk)
    assert len(res) == topk
    assert all("hello" in record.text for record in res)


@pytest.mark.db
@pytest.mark.parametrize("registry", [(JsonTable,)], indirect=True)
async def test_jsonb(registry):
    num = 10
    for i in range(num):
        await registry.insert(JsonTable(text=f"hello {i}", data=Jsonb({"key": i})))

    inserted = await registry.select_by(JsonTable.partial_init(), fields=["text"])
    assert len(inserted) == num


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Document, Chunk)], indirect=True)
async def test_foreign_key(registry):
    docs = [
        Document(text="hello world"),
        Document(text="hello there"),
    ]
    chunks = [
        Chunk(doc_id=1, text="hello", keyword=Keyword("hello"), vector=gen_vector()),
        Chunk(doc_id=1, text="world", keyword=Keyword("world"), vector=gen_vector()),
        Chunk(doc_id=1, text="no keyword field", keyword=None, vector=gen_vector()),
    ]
    for record in docs + chunks:
        await registry.insert(record)

    assert len(await registry.select_by(Document.partial_init())) == len(docs)
    assert len(await registry.select_by(Chunk.partial_init())) == len(chunks)
    # remove the doc should also remove the related chunks
    await registry.remove_by(Document.partial_init(uid=1))
    assert len(await registry.select_by(Document.partial_init())) == 1
    assert len(await registry.select_by(Chunk.partial_init())) == 0


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Document, Chunk)], indirect=True)
async def test_injection(registry):
    @registry.inject(output=Document)
    def create_doc(text: str) -> Document:
        return Document(text=text)

    @registry.inject(input=Document, output=Chunk)
    async def create_chunk(uid: int, text: str) -> AsyncGenerator[Chunk]:
        for t in text.split():
            yield Chunk(doc_id=uid, text=t, keyword=Keyword(t), vector=gen_vector())

    text = "hello world what happened to vector search"
    await create_doc(text)
    await create_chunk()

    docs = await registry.select_by(Document.partial_init())
    assert len(docs) == 1

    chunks = await registry.select_by(Chunk.partial_init())
    assert len(chunks) == len(text.split())

    topk = 3
    # vector search
    vec_res = await registry.search_by_vector(Chunk, gen_vector(), topk=topk)
    assert len(vec_res) == topk
    assert all(chunk.text in text for chunk in vec_res)
    # keyword search
    text_res = await registry.search_by_keyword(Chunk, "vector", topk=topk)
    assert len(text_res) == 1


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Sentence,)], indirect=True)
async def test_multi_vec_maxsim(registry):
    @registry.inject(output=Sentence)
    def create_sentence(text: str) -> Sentence:
        return Sentence(
            text=text, vector=[gen_vector() for _ in range(len(text.split()))]
        )

    text = "the quick brown fox jumps over the lazy dog"
    num = 32
    for _ in range(num):
        await create_sentence(text)
    sentence = await registry.select_by(Sentence.partial_init())
    assert len(sentence) == num
    assert len(sentence[0].vector) == len(text.split())

    topk = 3
    for dim in range(1, 10):
        res = await registry.search_by_multivec(
            Sentence, [gen_vector() for _ in range(dim)], topk=topk
        )
        assert len(res) == topk


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Document, Chunk)], indirect=True)
async def test_pipeline(registry):
    @registry.inject(output=Document)
    def create_doc(text: str) -> Document:
        return Document(text=text)

    @registry.inject(input=Document, output=Chunk)
    def create_chunk(uid: int, text: str) -> list[Chunk]:
        nums = [int(x) for x in text.split()]
        return [
            Chunk(
                doc_id=uid,
                text=f"num[{num}]",
                keyword=Keyword(num),
                vector=gen_vector(),
            )
            for num in nums
        ]

    correct = "1 2 3 4 5"
    error = "100 0.1 no no"
    pipeline = registry.create_pipeline([create_doc, create_chunk])

    await pipeline.run(correct)
    docs = await registry.select_by(Document.partial_init())
    assert len(docs) == 1
    chunks = await registry.select_by(Chunk.partial_init())
    assert len(chunks) == len(correct.split())

    # break the transaction won't add new records
    with pytest.raises(ValueError):
        await pipeline.run(error)

    docs = await registry.select_by(Document.partial_init())
    assert len(docs) == 1
    chunks = await registry.select_by(Chunk.partial_init())
    assert len(chunks) == len(correct.split())


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Image,)], indirect=True)
async def test_multivec_copy(registry):
    contents = (
        "hello world",
        "hello there",
        "hi there, how are you",
        "the quick brown fox jumps over the lazy dog",
        "Across the Great Wall we can reach every corner in the world",
    )
    await registry.copy_bulk(
        [
            Image(text=text, vecs=[gen_vector() for _ in range(len(text.split()))])
            for text in contents
        ]
    )

    assert len(await registry.select_by(Image.partial_init())) == len(contents)

    await registry.copy_bulk(
        [
            Image(
                text="not bad as a test case",
                vecs=[gen_vector() for _ in range(len(contents))],
                dataset="test",
                oid=0,
            )
        ]
    )
    assert len(await registry.select_by(Image.partial_init())) == len(contents) + 1


@pytest.mark.db
@pytest.mark.parametrize("registry", [(Document, Chunk)], indirect=True)
async def test_search_return(registry):
    num = 100
    topk = 5
    await registry.insert(Document(text="hello world"))
    chunks = []
    for i in range(num):
        text = f"hello {i}"
        chunks.append(
            Chunk(doc_id=1, text=text, vector=gen_vector(), keyword=Keyword(text))
        )
    await asyncio.gather(*[registry.insert(chunk) for chunk in chunks])

    inserted: list[Chunk] = await registry.select_by(
        Chunk.partial_init(), fields=["text"]
    )
    assert len(inserted) == num
    for record in inserted:
        assert record.text.startswith("hello")
        # vector field is not selected by default
        assert record.vector is msgspec.UNSET
        assert record.keyword is msgspec.UNSET

    res = await registry.search_by_vector(Chunk, gen_vector(), topk=topk)
    assert len(res) == topk
    assert all(record.vector is msgspec.UNSET for record in res)

    res = await registry.search_by_keyword(Chunk, "hello", topk=topk)
    assert len(res) == topk
    assert all(record.keyword is msgspec.UNSET for record in res)



from datetime import datetime

import msgspec


# `frozen=True` is required to make it hashable
class GraphEntity(msgspec.Struct, kw_only=True, frozen=True):
    text: str
    label: str = ""
    description: str = ""


class GraphRelation(msgspec.Struct, kw_only=True, frozen=True):
    source: GraphEntity
    target: GraphEntity
    description: str = ""


class Document(msgspec.Struct, kw_only=True):
    ext: str
    data: bytes
    path: str = ""
    digest: str = ""
    source: str = ""
    updated_at: datetime = msgspec.field(default_factory=datetime.now)


class SparseEmbedding(msgspec.Struct, kw_only=True, frozen=True):
    dim: int
    indices: list[int]
    values: list[float]


class RetrievedChunk(msgspec.Struct, kw_only=True):
    uid: str
    text: str
    score: float



class VechordError(Exception):
    """Base class for all Vechord exceptions."""


class APIKeyUnsetError(VechordError):
    """Raised when the API key is not set."""

    def __init__(self, name: str):
        self.message = (
            f"API key `{name}` is not set. Please set the environment variable."
        )
        super().__init__(self.message)


class HTTPCallError(VechordError):
    def __init__(
        self, msg: str, status_code: int = 500, err_msg: str = "internal error"
    ):
        self.message = f"{msg} [{status_code}]: {err_msg}"
        super().__init__(self.message)

    def __str__(self):
        return "Failed to call external LLM/Embedding services"


class DecodeStructuredOutputError(VechordError):
    """Raised when decoding structured output fails."""

    def __str__(self):
        return "Failed to decode structured output from LLM services"


class UnexpectedResponseError(VechordError):
    """Raised when the HTTP response is not as expected."""

    def __str__(self):
        return "Unexpected response from LLM/Embedding services"


class RequestError(VechordError):
    """Raised with bad request."""


class TimeoutError(VechordError):
    """Raised when a request times out."""


def extract_safe_err_msg(exc: Exception) -> str:
    if not isinstance(exc, VechordError):
        return ""
    return str(exc)



from vechord.augment import GeminiAugmenter
from vechord.chunk import GeminiChunker, RegexChunker, SpacyChunker
from vechord.client import VechordClient
from vechord.embedding import (
    GeminiDenseEmbedding,
    OpenAIDenseEmbedding,
    SpacyDenseEmbedding,
)
from vechord.evaluate import GeminiEvaluator
from vechord.extract import GeminiExtractor, SimpleExtractor
from vechord.load import LocalLoader
from vechord.model import Document
from vechord.registry import VechordPipeline, VechordRegistry
from vechord.rerank import CohereReranker
from vechord.service import create_web_app
from vechord.spec import (
    DefaultDocument,
    ForeignKey,
    IndexColumn,
    Keyword,
    KeywordIndex,
    MultiVectorIndex,
    PrimaryKeyAutoIncrease,
    PrimaryKeyUUID,
    Table,
    Vector,
    VectorIndex,
    create_chunk_with_dim,
)

__all__ = [
    "CohereReranker",
    "DefaultDocument",
    "Document",
    "ForeignKey",
    "GeminiAugmenter",
    "GeminiChunker",
    "GeminiDenseEmbedding",
    "GeminiEvaluator",
    "GeminiExtractor",
    "IndexColumn",
    "Keyword",
    "KeywordIndex",
    "LocalLoader",
    "MultiVectorIndex",
    "OpenAIDenseEmbedding",
    "PrimaryKeyAutoIncrease",
    "PrimaryKeyUUID",
    "RegexChunker",
    "SimpleExtractor",
    "SpacyChunker",
    "SpacyDenseEmbedding",
    "Table",
    "VechordClient",
    "VechordPipeline",
    "VechordRegistry",
    "Vector",
    "VectorIndex",
    "create_chunk_with_dim",
    "create_web_app",
]



from datetime import datetime, timezone
from functools import partial
from typing import Annotated
from uuid import UUID

import msgspec
import numpy as np
import pytest

from tests.conftest import gen_vector
from vechord.service import vechord_decode_hook, vechord_encode_hook
from vechord.spec import (
    DefaultDocument,
    ForeignKey,
    Keyword,
    MultiVectorIndex,
    PrimaryKeyAutoIncrease,
    PrimaryKeyUUID,
    Table,
    Vector,
    VectorDistance,
    VectorIndex,
    create_chunk_with_dim,
)


class Document(Table, kw_only=True):
    uid: PrimaryKeyAutoIncrease | None = None
    title: str
    text: str
    updated_at: datetime = msgspec.field(
        default_factory=partial(datetime.now, timezone.utc)
    )


class Chunk(Table, kw_only=True):
    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    doc_id: Annotated[int, ForeignKey[Document.uid]]
    text: str
    vec: Vector[128]
    multivec: list[Vector[128]]
    keyword: Keyword


class Simple(Table):
    uid: int
    text: str


def test_table_serde():
    chunk = Chunk(
        doc_id=1,
        text="hello",
        vec=gen_vector(),
        multivec=[gen_vector(), gen_vector()],
        keyword=Keyword("hello"),
    )
    encoded = msgspec.json.encode(chunk, enc_hook=vechord_encode_hook)
    decoded = msgspec.json.decode(encoded, type=Chunk, dec_hook=vechord_decode_hook)
    assert isinstance(decoded, Chunk)


@pytest.mark.parametrize(
    "table", [Document, Chunk, Simple, DefaultDocument, create_chunk_with_dim(128)]
)
def test_storage_cls_methods(table: type[Table]):
    assert table.name() == table.__name__.lower()
    assert "uid" in table.fields()

    t = table.partial_init()
    for field in t.fields():
        assert getattr(t, field) is msgspec.UNSET

    # UNSET won't appear in the `todict` result
    assert t.todict() == {}


def test_table_todict():
    doc = Document(title="hello world", text="hello there")
    values = doc.todict()
    assert values["title"] == "hello world"
    assert values["text"] == "hello there"
    assert values["updated_at"] is not None, values
    assert "uid" not in values, values

    chunk = Chunk(
        doc_id=1,
        text="hello",
        vec=Vector[128]([0.1] * 128),
        multivec=[],
        keyword=Keyword("hello"),
    )
    values = chunk.todict()
    assert values["doc_id"] == 1
    assert values["text"] == "hello"
    assert np.isclose(values["vec"], np.array([0.1] * 128)).all()
    assert values["multivec"] == []
    assert values["keyword"] == "hello"
    assert values["uid"], values


def test_default_chunk():
    dim = 1024
    chunk_cls = create_chunk_with_dim(dim)
    DenseVector = Vector[dim]
    text = "hello world"
    chunk = chunk_cls(
        doc_id=0, text=text, vec=DenseVector(np.ones(dim)), keyword=Keyword(text)
    )

    assert isinstance(chunk.uid, UUID)
    values = chunk.todict()
    assert values["doc_id"] == 0


def test_table_cls_methods():
    assert Document.primary_key() == "uid", Document
    assert Chunk.primary_key() == "uid", Chunk

    assert Document.vector_column() is None
    assert Chunk.vector_column().name == "vec"
    assert Chunk.multivec_column().name == "multivec"
    assert Chunk.keyword_column().name == "keyword"

    def find_schema_by_name(schema, name):
        for n, t in schema:
            if n == name:
                return t

    assert "GENERATED ALWAYS AS IDENTITY" in find_schema_by_name(
        Document.table_schema(), "uid"
    )
    assert "VECTOR(128)" in find_schema_by_name(Chunk.table_schema(), "vec")
    assert (
        "REFERENCES {namespace}_document(uid) ON DELETE CASCADE"
        in find_schema_by_name(Chunk.table_schema(), "doc_id")
    )

    assert find_schema_by_name(Chunk.table_psql_types(), "vec") == "vector"
    assert find_schema_by_name(Chunk.table_psql_types(), "multivec") == "vector[]"
    assert find_schema_by_name(Chunk.table_psql_types(), "keyword") == "bm25vector"
    assert find_schema_by_name(Chunk.table_psql_types(), "uid") == "uuid"

    assert Document.non_vec_columns() == Document.fields()
    assert Chunk.non_vec_columns() == ("uid", "doc_id", "text")


def test_vector_type():
    Dense = Vector[128]

    # test the dim
    with pytest.raises(ValueError):
        Dense([0.1] * 100)

    with pytest.raises(ValueError):
        Dense(np.random.rand(123))

    assert np.equal(Dense(np.ones(128)), Dense([1.0] * 128)).all()


def test_index():
    with pytest.raises(msgspec.ValidationError):
        VectorIndex(distance="bug")

    with pytest.raises(msgspec.ValidationError):
        VectorIndex(lists="bug")

    lists = 128
    vec_idx = VectorIndex(distance="l2", lists=lists)
    assert vec_idx.distance is VectorDistance.L2
    assert vec_idx.lists == lists
    assert vec_idx.op_symbol == "<->"
    assert str(vec_idx.lists) in vec_idx.config()

    multivec_idx = MultiVectorIndex(lists=1)
    assert multivec_idx.lists == 1
    index_config = MultiVectorIndex().config()
    assert "spherical_centroids = true" in index_config
    assert "lists = []" in index_config


def test_annotated_index():
    lists = 8

    class Sentence(Table, kw_only=True):
        uid: PrimaryKeyAutoIncrease | None = None
        text: str
        vec: Annotated[Vector[128], VectorIndex(distance="cos", lists=lists)]
        vecs: Annotated[list[Vector[128]], MultiVectorIndex(lists=lists)]

    vec_col = Sentence.vector_column()
    assert vec_col
    assert vec_col.name == "vec"
    assert vec_col.index.op_name == "vector_cosine_ops"
    assert vec_col.index.lists == lists

    multivec_col = Sentence.multivec_column()
    assert multivec_col
    assert multivec_col.name == "vecs"
    assert multivec_col.index.lists == lists



from abc import ABC, abstractmethod

import msgspec

from vechord.errors import DecodeStructuredOutputError
from vechord.model import (
    GeminiGenerateRequest,
    GeminiMimeType,
    GraphEntity,
    GraphRelation,
)
from vechord.provider import GeminiGenerateProvider


class BaseEntityRecognizer(ABC):
    @abstractmethod
    def recognize(self, text: str) -> list[GraphEntity]:
        """Predict the entities in the text."""
        raise NotImplementedError

    @abstractmethod
    def recognize_with_relations(
        self, text: str
    ) -> tuple[list[GraphEntity], list[GraphRelation]]:
        """Predict the entities and relations in the text."""
        raise NotImplementedError


class SpacyEntityRecognizer(BaseEntityRecognizer):
    """Spacy Entity Recognizer."""

    def __init__(self, model: str = "en_core_web_sm"):
        import spacy

        self.nlp = spacy.load(model)
        self.matcher = spacy.matcher.Matcher(self.nlp.vocab)
        self.matcher.add(
            "ENT_VERB_ENT",
            [
                [
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                    {"POS": "VERB", "OP": "+"},
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                ]
            ],
        )
        self.matcher.add(
            "ENT_PREP_ENT",
            [
                [
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                    {"POS": "AUX", "OP": "*"},
                    {"POS": "VERB"},
                    {"POS": "ADP"},
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                ]
            ],
        )
        self.matcher.add(
            "ENT_POSSESSIVE_ENT",
            [
                [
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                    {"IS_PUNCT": True, "OP": "?"},
                    {"LOWER": {"IN": ["'s"]}, "OP": "?"},
                    {"POS": "NOUN"},
                    {"LOWER": "is", "OP": "?"},
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                ]
            ],
        )
        self.matcher.add(
            "ENT_APPOSITION_ENT",
            [
                [
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                    {"IS_PUNCT": True, "OP": "?"},
                    {"POS": "NOUN", "OP": "+"},
                    {"LOWER": "of", "OP": "?"},
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                ]
            ],
        )
        self.matcher.add(
            "ENT_ATTRIBUTE_ENT",
            [
                [
                    {"ENT_TYPE": {"NOT_IN": [""]}},
                    {"IS_PUNCT": True, "OP": "?"},
                    {"POS": "NOUN"},
                    {"LIKE_NUM": True},
                ]
            ],
        )
        self.model = model

    def recognize(self, text) -> list[GraphEntity]:
        doc = self.nlp(text)
        return [GraphEntity(text=ent.text, label=ent.label_) for ent in doc.ents]

    def recognize_with_relations(
        self, text
    ) -> tuple[list[GraphEntity], list[GraphRelation]]:
        doc = self.nlp(text)
        ents = [GraphEntity(text=ent.text, label=ent.label_) for ent in doc.ents]
        relations: list[GraphRelation] = []
        matches = self.matcher(doc)
        for _, start, end in matches:
            span = doc[start:end]
            ent0 = ent1 = None
            for token in span:
                if token.ent_type_:
                    ent = GraphEntity(text=token.text, label=token.ent_type_)
                    if ent0 is None:
                        ent0 = ent
                    else:
                        ent1 = ent

            relations.append(
                GraphRelation(
                    source=ent0
                    or GraphEntity(text=span[0].text, label=span[0].ent_type_),
                    target=ent1
                    or GraphEntity(text=span[-1].text, label=span[-1].ent_type_),
                    description=" ".join(token.text for token in span),
                )
            )

        return ents, relations


class GeminiEntityRecognizer(BaseEntityRecognizer, GeminiGenerateProvider):
    """Entity recognizer using Gemini API.

    Since Gemini SDK doesn't support passing JSON schema directly, we have to
    build the HTTP request manually. Otherwise, we can only use pydantic models
    to define the returned types, which is not ideal for this use case. Internally,
    Gemini also generates the JSON schema from the pydantic model.
    """

    def __init__(self, model: str = "gemini-2.5-flash"):
        super().__init__(model)

    async def recognize(self, text) -> list[GraphEntity]:
        prompt = (
            "Given the text document, identify and extract all the entities, return the JSON "
            "format with the following fields: "
            "- text: the entity text "
            "- label: the entity type (e.g., PER, ORG, LOC, TIME, GPE, VEH etc.) "
            "- description: a brief description of the entity in the current context "
            "\n<document>\n{text}\n</document>\n"
        )
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_structure_response(
                prompt=prompt.format(text),
                schema=msgspec.json.schema(list[GraphEntity]),
            )
        )
        try:
            ents = msgspec.json.decode(resp.get_text(), type=list[GraphEntity])
        except (msgspec.DecodeError, KeyError) as err:
            raise DecodeStructuredOutputError(
                "Failed to decode Gemini response"
            ) from err
        return ents

    def decode_relations(
        self, text: str
    ) -> tuple[list[GraphEntity], list[GraphRelation]]:
        try:
            rels = msgspec.json.decode(text, type=list[GraphRelation])
        except (msgspec.DecodeError, KeyError) as err:
            raise DecodeStructuredOutputError(
                "Failed to decode Gemini response"
            ) from err
        ents: dict[str, GraphEntity] = {}
        for rel in rels:
            ents[rel.source.text] = rel.source
            ents[rel.target.text] = rel.target
        return list(ents.values()), rels

    async def recognize_image(
        self, img: bytes
    ) -> tuple[list[GraphEntity], list[GraphRelation]]:
        """Recognize entities & relations from the image."""
        prompt = (
            "Given the image, extract the meaningful entities and their relations "
            "between them. Entity could be person, location, org, event or category. "
            "Return a list of relations with source and target entities like: "
            "- source: the source entity (text, label and description) "
            "- target: the target entity (text, label and description) "
            "- description: a brief description of the relation in the current context "
        )
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_data_structure_resp(
                prompt=prompt,
                mime_type=GeminiMimeType.JPEG,
                data=img,
                schema=msgspec.json.schema(list[GraphRelation]),
            )
        )
        return self.decode_relations(resp.get_text())

    async def recognize_with_relations(
        self, text
    ) -> tuple[list[GraphEntity], list[GraphRelation]]:
        prompt = (
            "Given the text document, extract entities and the possible relations "
            "between them. Entity could be person, location, org, event or category. "
            "Return a list of relations with source and target entities like: "
            "- source: the source entity (text, label and description) "
            "- target: the target entity (text, label and description) "
            "- description: a brief description of the relation in the current context "
            "\n<document>\n{text}\n</document>\n"
        )
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_structure_response(
                prompt=prompt.format(text=text),
                schema=msgspec.json.schema(list[GraphRelation]),
            )
        )
        return self.decode_relations(resp.get_text())


if __name__ == "__main__":

    async def main():
        async with GeminiEntityRecognizer() as recognizer:
            text = "Barack Obama was the 44th President of the United States. He was born in Hawaii. His wife, Michelle Obama, is a lawyer and writer."

            ents, rels = await recognizer.recognize_with_relations(text)
            print(ents)
            print(rels)

    import asyncio

    asyncio.run(main())



from vechord.model.gemini import (
    GeminiEmbeddingRequest,
    GeminiEmbeddingResponse,
    GeminiGenerateRequest,
    GeminiGenerateResponse,
    GeminiMimeType,
    UMBRELAScore,
)
from vechord.model.internal import (
    Document,
    GraphEntity,
    GraphRelation,
    RetrievedChunk,
    SparseEmbedding,
)
from vechord.model.jina import JinaEmbeddingRequest, JinaEmbeddingResponse
from vechord.model.llamacloud import (
    LlamaCloudMimeType,
    LlamaCloudParseRequest,
    LlamaCloudParseResponse,
)
from vechord.model.voyage import (
    VoyageEmbeddingRequest,
    VoyageEmbeddingResponse,
    VoyageMultiModalEmbeddingRequest,
)
from vechord.model.web import InputType, ResourceRequest, RunAck, RunRequest

__all__ = [
    "Document",
    "GeminiEmbeddingRequest",
    "GeminiEmbeddingResponse",
    "GeminiGenerateRequest",
    "GeminiGenerateResponse",
    "GeminiMimeType",
    "GraphEntity",
    "GraphRelation",
    "InputType",
    "JinaEmbeddingRequest",
    "JinaEmbeddingResponse",
    "LlamaCloudMimeType",
    "LlamaCloudParseRequest",
    "LlamaCloudParseResponse",
    "ResourceRequest",
    "RetrievedChunk",
    "RunAck",
    "RunRequest",
    "SparseEmbedding",
    "UMBRELAScore",
    "VoyageEmbeddingRequest",
    "VoyageEmbeddingResponse",
    "VoyageMultiModalEmbeddingRequest",
]



from enum import Enum
from typing import Annotated, Any, Literal, Optional

import msgspec
import numpy as np

from vechord.typing import Self


# https://ai.google.dev/gemini-api/docs/document-processing#technical-details
# https://ai.google.dev/gemini-api/docs/image-understanding#supported-formats
class GeminiMimeType(str, Enum):
    # documents
    PDF = "application/pdf"
    JavaScript = "application/x-javascript"
    Python = "application/x-python"
    HTML = "text/html"
    CSS = "text/css"
    Markdown = "text/markdown"
    CSV = "text/csv"
    XML = "text/xml"
    RTF = "text/rtf"
    # images
    PNG = "image/png"
    JPEG = "image/jpeg"
    WEBP = "image/webp"
    HEIC = "image/heic"
    HEIF = "image/heif"


class UMBRELAScore(msgspec.Struct, kw_only=True):
    """Score for UMBRELA evaluation.

    - 0: Not relevant
    - 1: Relevant but does not answer the query
    - 2: Answer the query but may be a bit unclear
    - 3: Dedicated to the query and contains the exact answer
    """

    score: Annotated[int, msgspec.Meta(ge=0, le=3)]


class InlineData(msgspec.Struct, kw_only=True):
    mime_type: GeminiMimeType
    data: bytes


class ContentPart(msgspec.Struct, kw_only=True, omit_defaults=True):
    text: str | None = None
    inline_data: InlineData | None = None


class Part(msgspec.Struct):
    parts: list[ContentPart] = msgspec.field(default_factory=list)


class GenerationConfig(msgspec.Struct, kw_only=True):
    response_mime_type: Literal["application/json"] = "application/json"
    response_json_schema: dict[str, Any] | None = None


class GeminiGenerateRequest(msgspec.Struct, kw_only=True):
    contents: Part
    generation_config: Optional[GenerationConfig] = msgspec.field(
        default=None, name="generationConfig"
    )

    @classmethod
    def from_prompt_with_data(
        cls, prompt: str, mime_type: GeminiMimeType, data: bytes
    ) -> Self:
        return GeminiGenerateRequest(
            contents=Part(
                [
                    ContentPart(text=prompt),
                    ContentPart(inline_data=InlineData(mime_type=mime_type, data=data)),
                ]
            )
        )

    @classmethod
    def from_prompt_data_structure_resp(
        cls, prompt: str, mime_type: GeminiMimeType, data: bytes, schema: dict[str, Any]
    ) -> Self:
        return GeminiGenerateRequest(
            contents=Part(
                [
                    ContentPart(text=prompt),
                    ContentPart(inline_data=InlineData(mime_type=mime_type, data=data)),
                ]
            ),
            generation_config=GenerationConfig(response_json_schema=schema),
        )

    @classmethod
    def from_prompt(cls, prompt: str) -> Self:
        return GeminiGenerateRequest(contents=Part(parts=[ContentPart(text=prompt)]))

    @classmethod
    def from_prompt_structure_response(
        cls, prompt: str, schema: dict[str, Any]
    ) -> Self:
        return GeminiGenerateRequest(
            contents=Part(parts=[ContentPart(text=prompt)]),
            generation_config=GenerationConfig(response_json_schema=schema),
        )


class ContentResponse(msgspec.Struct, kw_only=True, omit_defaults=True):
    content: Part


class GeminiGenerateResponse(msgspec.Struct, kw_only=True):
    candidates: list[ContentResponse] = msgspec.field(default_factory=list)

    def get_text(self) -> str:
        """Get the text from the first candidate's content part."""
        if not self.candidates or not self.candidates[0].content.parts:
            return ""
        return self.candidates[0].content.parts[0].text or ""


# https://ai.google.dev/gemini-api/docs/embeddings#supported-task-types
GeminiEmbeddingType = Literal[
    "SEMANTIC_SIMILARITY",
    "CLASSIFICATION",
    "CLUSTERING",
    "CODE_RETRIEVAL_QUERY",
    "RETRIEVAL_DOCUMENT",
    "RETRIEVAL_QUERY",
    "QUESTION_ANSWERING",
    "FACT_VERIFICATION",
]


class GeminiEmbeddingRequest(msgspec.Struct, kw_only=True, omit_defaults=True):
    model: Optional[str] = None
    task_type: GeminiEmbeddingType = msgspec.field(
        default="SEMANTIC_SIMILARITY", name="taskType"
    )
    content: Part

    @classmethod
    def from_text_with_type(
        cls, text: str, task_type: GeminiEmbeddingType = "SEMANTIC_SIMILARITY"
    ) -> Self:
        return GeminiEmbeddingRequest(
            content=Part(parts=[ContentPart(text=text)]),
            task_type=task_type,
        )


class Embedding(msgspec.Struct):
    values: list[float]


class GeminiEmbeddingResponse(msgspec.Struct, kw_only=True):
    embedding: Embedding

    def get_emb(self) -> np.ndarray:
        """Get the first embedding as a numpy array."""
        return np.array(self.embedding.values, dtype=np.float32)



import asyncio
from collections.abc import AsyncIterator
from http import HTTPStatus
from uuid import UUID

import msgspec
import pytest
from falcon import testing

from tests.conftest import gen_vector
from vechord.service import create_web_app
from vechord.spec import DefaultDocument, Keyword, create_chunk_with_dim

pytestmark = pytest.mark.anyio
DefaultChunk = create_chunk_with_dim(128)


@pytest.fixture(name="client")
async def fixture_client(registry):
    @registry.inject(output=DefaultDocument)
    def create_doc(text: str) -> DefaultDocument:
        return DefaultDocument(text=text)

    @registry.inject(input=DefaultDocument, output=DefaultChunk)
    async def create_chunk(uid: UUID, text: str) -> AsyncIterator[DefaultChunk]:
        nums = [int(x) for x in text.split()]
        for num in nums:
            yield DefaultChunk(
                doc_id=uid,
                text=f"num[{num}]",
                keyword=Keyword(num),
                vec=gen_vector(),
            )

    app = create_web_app(registry, registry.create_pipeline([create_doc, create_chunk]))

    async with testing.TestClient(app) as client:
        yield client


@pytest.mark.parametrize("registry", [(DefaultDocument, DefaultChunk)], indirect=True)
async def test_service_health(client):
    resp = await client.simulate_get("/")
    assert resp.status_code == HTTPStatus.OK


@pytest.mark.parametrize("registry", [(DefaultDocument, DefaultChunk)], indirect=True)
async def test_service_openapi(client):
    resp = await client.simulate_get("/openapi/spec.json")
    assert resp.status_code == HTTPStatus.OK
    assert resp.json is not None

    resp = await client.simulate_get("/openapi/swagger")
    assert resp.status_code == HTTPStatus.OK


@pytest.mark.parametrize("registry", [(DefaultDocument, DefaultChunk)], indirect=True)
async def test_service_table(client):
    resp = await client.simulate_get("/api/table/defaultdocument")
    assert resp.status_code == HTTPStatus.OK
    assert len(resp.json) == 0

    text = "hello world"

    doc = DefaultDocument(text=text)
    resp = await client.simulate_post(
        "/api/table/defaultdocument", body=msgspec.json.encode(doc)
    )
    assert resp.status_code == HTTPStatus.CREATED

    resp = await client.simulate_get("/api/table/defaultdocument")
    assert resp.status_code == HTTPStatus.OK
    assert len(resp.json) == 1
    assert resp.json[0]["text"] == text

    resp = await client.simulate_delete(
        "/api/table/defaultdocument", json={"text": text}
    )
    assert resp.status_code == HTTPStatus.OK

    resp = await client.simulate_get("/api/table/defaultdocument")
    assert resp.status_code == HTTPStatus.OK
    assert len(resp.json) == 0


@pytest.mark.parametrize("registry", [(DefaultDocument, DefaultChunk)], indirect=True)
async def test_service_pipeline(client):
    text = "1 2 3 4 5"
    resp = await client.simulate_post("/api/pipeline", json={"text": text})
    assert resp.status_code == HTTPStatus.OK

    resp = await client.simulate_get("/api/table/defaultchunk")
    assert resp.status_code == HTTPStatus.OK
    chunks = resp.json
    assert len(chunks) == len(text.split())
    for i, chunk in enumerate(chunks):
        assert chunk["text"] == f"num[{i + 1}]"


@pytest.mark.parametrize("registry", [(DefaultDocument, DefaultChunk)], indirect=True)
async def test_concurrent_db_transaction(client):
    requests = [
        client.simulate_post(
            "/api/pipeline", json={"text": " ".join(map(str, range(i, i + 5)))}
        )
        for i in range(5)
    ]
    responses = await asyncio.gather(*requests)
    for resp in responses:
        assert resp.status_code == HTTPStatus.OK, resp.content



from typing import Literal, Optional

import msgspec
import numpy as np

from vechord.errors import RequestError, UnexpectedResponseError
from vechord.typing import Self

# https://jina.ai/api-dashboard/embedding
JinaEmbeddingType = Literal[
    "retrieval.query",
    "retrieval.passage",
    "code.query",
    "code.passage",
    "text-matching",
]


class JinaInput(msgspec.Struct, kw_only=True, omit_defaults=True):
    text: Optional[str] = None
    image: Optional[str | bytes] = None  # URL or base64 encoded image
    pdf: Optional[str] = None  # URL or base64 encoded PDF


class JinaEmbeddingRequest(msgspec.Struct, kw_only=True, omit_defaults=True):
    model: Literal["jina-embeddings-v4", "jina-embeddings-v3"]
    dimensions: int = 2048
    truncate: bool
    task: JinaEmbeddingType
    embedding_type: Literal["binary", "ubinary", "base64", "float"] = "float"
    input_content: list[JinaInput] = msgspec.field(name="input")

    @classmethod
    def from_text(cls, text: str, task: JinaEmbeddingType, model: str) -> Self:
        return JinaEmbeddingRequest(
            model=model,
            truncate=True,
            task=task,
            embedding_type="base64",
            input_content=[JinaInput(text=text)],
        )

    @classmethod
    def from_text_image(
        cls,
        text: str,
        image: bytes,
        image_url: str,
        task: JinaEmbeddingType,
        model: str,
    ) -> Self:
        req = JinaEmbeddingRequest(
            model=model,
            truncate=True,
            task=task,
            embedding_type="base64",
            input_content=[],
        )
        if not (text or image or image_url):
            raise RequestError("At least one of text, image must be provided")
        if text:
            req.input_content.append(JinaInput(text=text))
        if image:
            req.input_content.append(JinaInput(image=image))
        if image_url:
            req.input_content.append(JinaInput(image=image_url))
        return req


class EmbeddingObject(msgspec.Struct, kw_only=True, omit_defaults=True):
    object: Literal["embedding"] = "embedding"
    index: int
    embedding: list[float] | bytes


class JinaEmbeddingResponse(msgspec.Struct, kw_only=True):
    data: list[EmbeddingObject]

    def get_emb(self) -> np.ndarray:
        """Get the first embedding as a numpy array."""
        if not self.data or not self.data[0].embedding:
            raise UnexpectedResponseError("empty embedding data")
        emb = self.data[0].embedding
        if isinstance(emb, list):
            return np.array(emb, dtype=np.float32)
        return np.frombuffer(emb, dtype=np.float32)



import unicodedata
from abc import ABC, abstractmethod
from html.parser import HTMLParser

import pypdfium2 as pdfium

from vechord.log import logger
from vechord.model import (
    Document,
    GeminiGenerateRequest,
    GeminiMimeType,
    LlamaCloudMimeType,
    LlamaCloudParseRequest,
)
from vechord.provider import GeminiGenerateProvider, LlamaCloudProvider


class BaseHTMLParser(HTMLParser):
    """A simple HTML parser to extract text content."""

    def __init__(self) -> None:
        super().__init__()
        self.content: list[str] = []
        self.skip = False

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        if tag in ("script", "style"):
            self.skip = True

    def handle_endtag(self, tag: str) -> None:
        if tag in ("script", "style"):
            self.skip = False

    def handle_data(self, data: str) -> None:
        if not self.skip:
            self.content.append(data.strip())


class BaseExtractor(ABC):
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    async def extract_pdf(self, doc: bytes) -> str:
        raise NotImplementedError

    @abstractmethod
    async def extract_html(self, text: str) -> str:
        raise NotImplementedError

    async def extract(self, doc: Document) -> str:
        if doc.ext == ".txt":
            text = doc.data.decode("utf-8")
        elif doc.ext == ".pdf":
            text = await self.extract_pdf(doc.data)
        elif doc.ext == ".html":
            text = await self.extract_html(doc.data.decode("utf-8"))
        else:
            logger.warning("unsupported file type '%s' for %s", doc.ext, doc.path)
            text = ""
        return unicodedata.normalize("NFKC", text)


class SimpleExtractor(BaseExtractor):
    """Local extractor for text files."""

    def name(self) -> str:
        return "basic_extractor"

    async def extract_pdf(self, doc: bytes) -> str:
        """Extract text from PDF using pypdfium2."""
        pdf = pdfium.PdfDocument(doc)
        text = []
        for page in pdf:
            text.append(page.get_textpage().get_text_bounded())

        return "\n".join(text)

    async def extract_html(self, text: str) -> str:
        """Extract text from HTML.

        Args:
            text: HTML text.
        """
        parser = BaseHTMLParser()
        parser.feed(text)
        return "\n".join(t for t in parser.content if t)


class GeminiExtractor(SimpleExtractor, GeminiGenerateProvider):
    """Extract text with Gemini model.

    Limits:
        - PDF: less than **20 MB** in size, max **3072x3072** with original aspect ratio
        - Image: less than **20 MB** in size, larger images are tiled into **768x768** tiles
    """

    def __init__(self, model: str = "gemini-2.5-flash"):
        super().__init__(model)
        self.pdf_prompt = (
            "Extract the main content from the PDF document. Ensure to exclude any "
            "metadata, headers, footers, or any other non-essential information. "
            "Return the extracted content as it appears in the document, without "
            "any additional modification, summarization or interpretation."
        )
        self.jpeg_prompt = (
            "Extract the visible text from the image, generate a concise caption "
            "describing the image's content or scene, return the text with caption."
        )

    def name(self) -> str:
        return f"gemini_extractor_{self.model}"

    async def extract_image(self, img: bytes) -> str:
        """Extract text & caption from image."""
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_with_data(
                self.jpeg_prompt, GeminiMimeType.JPEG, img
            )
        )
        return resp.get_text().strip()

    async def extract_pdf(self, doc: bytes) -> str:
        """Extract text from PDF page by page."""
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_with_data(
                self.pdf_prompt, GeminiMimeType.PDF, doc
            )
        )
        return resp.get_text().strip()


class LlamaParseExtractor(SimpleExtractor, LlamaCloudProvider):
    """Extract test with LlamaCloud Parse service.

    Limits:
        - Maximum run time for jobs : 30 minutes. If your job takes more than **30 minutes** to process, a TIMEOUT error will be raised.
        - Maximum size of files: **300Mb**.
        - Maximum image extracted / OCR per page: **35 images**. If more images are present in a page, only the 35 biggest one are extracted / OCR.
        - Maximum amount of text extracted per page: **64Kb**. Content beyond the 64Kb mark is ignored.
    """

    def __init__(self):
        super().__init__()

    def name(self) -> str:
        return "llama_parse_extractor"

    async def extract_image(self, img: bytes) -> str:
        """Extract text from image using LlamaCloud Parse service."""
        req = LlamaCloudParseRequest.from_image(img, LlamaCloudMimeType.JPEG)
        resp = await self.parse(req)
        text = await self.get_text(resp.id)
        if text is not None:
            return text.strip()
        return ""

    async def extract_pdf(self, doc: bytes) -> str:
        """Extract text from PDF using LlamaCloud Parse service."""
        req = LlamaCloudParseRequest.from_pdf(doc, LlamaCloudMimeType.PDF)
        resp = await self.parse(req)
        text = await self.get_text(resp.id)
        if text is not None:
            return text.strip()
        return ""



from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import AsyncExitStack
from os import environ
from typing import TypeVar

from vechord.spec import Table

T = TypeVar("T", bound=Table)


class BaseReranker(ABC):
    @abstractmethod
    async def rerank(self, query: str, chunks: list[str]) -> list[int]:
        """Return the indices of the reranked chunks."""
        raise NotImplementedError


class CohereReranker(BaseReranker):
    """Rerank chunks using Cohere API (requires env `COHERE_API_KEY`)."""

    def __init__(self, model: str = "rerank-v3.5"):
        self.api_key = environ.get("COHERE_API_KEY")
        if not self.api_key:
            raise ValueError("Cohere API key not found in environment variables.")

        import cohere

        self.client = cohere.AsyncClientV2(api_key=self.api_key)
        self.model = model

    async def __aenter__(self):
        self._async_exit_stack = AsyncExitStack()
        await self._async_exit_stack.enter_async_context(self.client)
        return self

    async def __aexit__(self, _exc_type, _exc_value, _traceback):
        await self._async_exit_stack.aclose()

    async def rerank(self, query: str, chunks: list[str]) -> list[int]:
        resp = await self.client.rerank(
            model=self.model,
            query=query,
            documents=chunks,
        )
        return [item.index for item in resp.results]


class ReciprocalRankFusion:
    """Fuse chunks using reciprocal rank."""

    def __init__(self, k: int = 60):
        self.k = k

    def get_score(self, rank: int) -> float:
        return 1 / (self.k + rank)

    def fuse(self, retrieved_chunks: list[list[T]]) -> list[T]:
        chunk_score: dict[str, float] = defaultdict(float)
        chunk_map: dict[str, Table] = {}
        for chunks in retrieved_chunks:
            for i, chunk in enumerate(chunks):
                chunk_score[getattr(chunk, chunk.primary_key())] += self.get_score(i)
                chunk_map[getattr(chunk, chunk.primary_key())] = chunk

        sorted_uid = sorted(chunk_score, key=lambda x: chunk_score[x], reverse=True)
        return [chunk_map[uid] for uid in sorted_uid]



from collections.abc import AsyncIterable, Iterable
from contextlib import AsyncExitStack
from functools import wraps
from inspect import isasyncgenfunction, iscoroutinefunction
from typing import (
    Callable,
    Optional,
    Sequence,
    TypeVar,
    get_origin,
    get_type_hints,
)

import numpy as np

from vechord.client import VechordClient, select_transaction_buffer_conn
from vechord.log import logger
from vechord.pipeline import VechordPipeline
from vechord.spec import Table, Vector

T = TypeVar("T", bound=Table)


def is_list_of_type(typ) -> bool:
    origin = get_origin(typ)
    if origin is None:
        return False
    return issubclass(origin, (Iterable, AsyncIterable))


def get_iterator_type(typ) -> type:
    if not is_list_of_type(typ):
        return typ
    return get_iterator_type(typ.__args__[0])


class VechordRegistry:
    """Create a registry for the given namespace and PostgreSQL URL.

    Args:
        namespace: the namespace for this registry, will be the prefix for all the
            tables registered.
        url: the PostgreSQL URL to connect to.
        tables: a list of Table classes to be registered.
        create_index: whether or not to create the index if not exists.
    """

    def __init__(
        self,
        namespace: str,
        url: str,
        *,
        tables: Iterable[type[Table]] = (),
        create_index: bool = True,
    ):
        self.ns = namespace
        self.client = VechordClient(namespace, url)
        self.tables: list[type[Table]] = []
        self.create_index = create_index
        self.pipeline: list[Callable] = []
        self._inited = False

        for table in tables:
            if not issubclass(table, Table):
                raise ValueError(f"unsupported class {table}")
            self.tables.append(table)

    async def __aenter__(self):
        if not self._inited:
            self._async_exit_stack = AsyncExitStack()
            await self._async_exit_stack.enter_async_context(self.client)
            await self.init_table_index()
            self._inited = True
        return self

    async def __aexit__(self, _exc_type, _exc_value, _traceback):
        await self._async_exit_stack.aclose()

    async def process_startup(self, scope, event):
        """Falcon ASGI middleware lifespan hook."""
        await self.__aenter__()

    async def process_shutdown(self, scope, event):
        """Falcon ASGI middleware lifespan hook."""
        await self.__aexit__(None, None, None)

    def reset_namespace(self, namespace: str):
        self.ns = namespace
        self.client.ns = namespace

    async def init_table_index(self, tables: Optional[Iterable[type[Table]]] = None):
        if tables is None:
            tables = self.tables

        for table in tables:
            await self.client.create_table_if_not_exists(
                table.name(), table.table_schema()
            )
            logger.debug("create table %s if not exists", table.name())

            if table.keyword_column() is not None:
                await self.client.create_tokenizer()

            if not self.create_index:
                continue

            # create index
            for index_column in (
                table.vector_column(),
                table.multivec_column(),
                table.keyword_column(),
                *table.unique_columns(),
            ):
                if index_column is None:
                    continue
                await self.client.create_index_if_not_exists(table.name(), index_column)
                logger.debug(
                    "create index for %s.%s if not exists",
                    table.name(),
                    index_column.name,
                )

    def create_pipeline(self, steps: list[Callable]) -> VechordPipeline:
        """Create the :class:`VechordPipeline` to run multiple functions in a transaction.

        Args:
            steps: a list of functions to be run in the pipeline.
        """
        return VechordPipeline(client=self.client, steps=steps)

    async def select_by(
        self,
        obj: T,
        fields: Optional[Sequence[str]] = None,
        limit: Optional[int] = None,
    ) -> list[T]:
        """Retrieve the requested fields for the given object stored in the DB.

        Args:
            obj: the object to be retrieved, this should be generated from
                :meth:`Table.partial_init`, while the given values will be used
                for filtering (``=`` or ``is``).
            fields: the fields to be retrieved, if not set, all the fields will be
                retrieved.
            limit: the maximum number of results to be returned, if not set, all
                the results will be returned.
        """
        if not isinstance(obj, Table):
            raise ValueError(f"unsupported class {type(obj)}")

        cls_fields = obj.fields()
        cls = obj.__class__
        if fields is not None:
            if any(f not in cls_fields for f in fields):
                raise ValueError(f"unknown fields {fields}")
        else:
            fields = cls_fields

        kvs = obj.todict()
        res = await self.client.select(cls.name(), fields, kvs, limit=limit)
        return [
            cls.partial_init(**{k: v for k, v in zip(fields, r, strict=False)})
            for r in res
        ]

    async def search_by_vector(
        self,
        cls: type[T],
        vec: np.ndarray | Vector,
        topk: int = 10,
        return_fields: Optional[Sequence[str]] = None,
        probe: Optional[int] = None,
    ) -> list[T]:
        """Search the vector for the given `Table` class.

        Args:
            cls: the `Table` class to be searched.
            vec: the vector to be searched.
            topk: the number of results to be returned.
            return_fields: the fields to be returned, if not set, all the
                non-[vector,keyword] fields will be returned.
            probe: how many K-means clusters to probe for the `vec`.
        """
        if not issubclass(cls, Table):
            raise ValueError(f"unsupported class {cls}")
        fields = return_fields or cls.non_vec_columns()
        vec_col = cls.vector_column()
        if vec_col is None:
            raise ValueError(f"no vector column found in {cls}")
        res = await self.client.query_vec(
            cls.name(),
            vec_col,
            vec,
            topk=topk,
            return_fields=fields,
            probe=probe,
        )
        return [
            cls.partial_init(**{k: v for k, v in zip(fields, r, strict=False)})
            for r in res
        ]

    async def search_by_multivec(  # noqa: PLR0913
        self,
        cls: type[T],
        multivec: np.ndarray,
        topk: int = 10,
        return_fields: Optional[Sequence[str]] = None,
        maxsim_refine: int = 1000,
        probe: Optional[int] = None,
    ) -> list[T]:
        """Search the multivec for the given `Table` class.

        Args:
            cls: the `Table` class to be searched.
            multivec: the multivec to be searched.
            topk: the number of results to be returned.
            maxsim_refine: the maximum number of document vectors to be compute with
                full-precision for each vector in the `multivec`. 0 means all the
                distances are compute with bit quantization.
            probe: how many K-means clusters to probe for each vector in the `multivec`.
            return_fields: the fields to be returned, if not set, all the
                non-[vector,keyword] fields will be returned.
        """
        if not issubclass(cls, Table):
            raise ValueError(f"unsupported class {cls}")
        fields = return_fields or cls.non_vec_columns()
        multivec_col = cls.multivec_column()
        if multivec_col is None:
            raise ValueError(f"no multivec column found in {cls}")
        res = await self.client.query_multivec(
            cls.name(),
            multivec_col,
            multivec,
            maxsim_refine=maxsim_refine,
            probe=probe,
            topk=topk,
            return_fields=fields,
        )
        return [
            cls.partial_init(**{k: v for k, v in zip(fields, r, strict=False)})
            for r in res
        ]

    async def search_by_keyword(
        self,
        cls: type[T],
        keyword: str,
        topk: int = 10,
        return_fields: Optional[Sequence[str]] = None,
    ) -> list[T]:
        """Search the keyword for the given `Table` class.

        Args:
            cls: the `Table` class to be searched.
            keyword: the keyword to be searched.
            topk: the number of results to be returned.
            return_fields: the fields to be returned, if not set, all the
                non-[vector,keyword] fields will be returned.
        """
        if not issubclass(cls, Table):
            raise ValueError(f"unsupported class {cls}")
        fields = return_fields or cls.non_vec_columns()
        keyword_col = cls.keyword_column()
        if keyword_col is None:
            raise ValueError(f"no keyword column found in {cls}")
        res = await self.client.query_keyword(
            cls.name(),
            keyword_col,
            keyword,
            topk=topk,
            return_fields=fields,
            tokenizer=cls.keyword_tokenizer(),
        )
        return [
            cls.partial_init(**{k: v for k, v in zip(fields, r, strict=False)})
            for r in res
        ]

    async def remove_by(self, obj: Table):
        """Remove the given object from the DB.

        Args:
            obj: the object to be removed, this should be a `Table.partial_init()`
                instance, which means given values will be used for filtering.
        """
        if not isinstance(obj, Table):
            raise ValueError(f"unsupported class {type(obj)}")

        kvs = obj.todict()
        await self.client.delete(obj.__class__.name(), kvs)

    async def insert(self, obj: Table):
        """Insert the given object to the DB.

        Args:
            obj: the object to be inserted
        """
        if not isinstance(obj, Table):
            raise ValueError(f"unsupported class {type(obj)}")
        await self.client.insert(obj.name(), obj.todict())

    async def copy_bulk(self, objs: list[Table]):
        """Insert the given list of objects to the DB.

        This is more efficient than calling `insert` for each object.

        Args:
            objs: the list of objects to be inserted, needs to be of the same
                class and filled with the same fields. The class should be a
                subclass of `Table`.
        """
        if not objs:
            return
        cls = objs[0].__class__
        if not issubclass(cls, Table):
            raise ValueError(f"unsupported class {cls}")
        if not all(isinstance(obj, cls) for obj in objs):
            raise ValueError(f"not all the objects are {cls}")
        if cls.keyword_column() is not None:
            raise RuntimeError("`copy_bulk` does not support keyword column")

        name = objs[0].name()
        values = [obj.todict() for obj in objs]
        keys = set(values[0].keys())
        types = tuple(v for k, v in objs[0].table_psql_types() if k in keys)
        await self.client.copy_bulk(name, values, types)

    def inject(
        self, input: Optional[type[Table]] = None, output: Optional[type[Table]] = None
    ):
        """Decorator to inject the data for the function arguments & return value.

        This can be applied to both sync & async functions. But the decorated
        functions will all be async functions.

        Args:
            input: the input table to be retrieved from the DB. If not set, the function
                will require the input to be passed in the function call.
            output: the output table to store the return value. If not set, the return
                value will be return to the caller in a `list`.
        """
        if input is None and output is None:
            return lambda func: func
        if input is not None and not issubclass(input, Table):
            raise ValueError(f"unsupported class {input}")
        if output is not None and not issubclass(output, Table):
            raise ValueError(f"unsupported class {output}")

        def decorator(func: Callable):
            hints = get_type_hints(func)
            returns = hints.pop("return", None)
            columns = hints.keys()
            return_type = get_iterator_type(returns)
            if output and return_type is not output:
                raise ValueError(
                    f"expected {output}, got {return_type} in {func} output"
                )

            if output is not None and returns is None:
                raise ValueError(
                    f"requires the return type for {func} if `output` is set"
                )

            _is_async = iscoroutinefunction(func)
            _is_async_gen = isasyncgenfunction(func)

            async def execute_func(*args, **kwargs):
                if _is_async:
                    return await func(*args, **kwargs)
                elif _is_async_gen:
                    # has to collect all items in an async generator
                    items = []
                    async for item in func(*args, **kwargs):
                        items.append(item)
                    return items
                return func(*args, **kwargs)

            @wraps(func)
            async def wrapper(*args, **kwargs):
                arguments = [args]
                if input is not None:
                    arguments = await self.client.select(
                        input.name(),
                        columns,
                        from_buffer=select_transaction_buffer_conn.get() is not None,
                    )

                if output is None:
                    return [await execute_func(*arg, **kwargs) for arg in arguments]

                count = 0
                use_copy = output.keyword_column() is None
                if is_list_of_type(returns):
                    for arg in arguments:
                        if use_copy:
                            rets = list(await execute_func(*arg, **kwargs))
                            await self.copy_bulk(rets)
                            count += len(rets)
                        else:
                            for ret in await execute_func(*arg, **kwargs):
                                await self.insert(ret)
                                count += 1
                elif use_copy:
                    # ref https://github.com/python/cpython/issues/76294
                    rets = [await execute_func(*arg, **kwargs) for arg in arguments]
                    await self.copy_bulk(rets)
                    count += len(rets)
                else:
                    for arg in arguments:
                        ret = await execute_func(*arg, **kwargs)
                        await self.insert(ret)
                        count += 1
                logger.debug("inserted %d items to %s", count, output.name())

            return wrapper

        return decorator

    async def clear_storage(self, drop_table: bool = False):
        """Clear the storage of the registry.

        Args:
            drop_table: whether to drop the table after removing all the data.
        """
        for table in self.tables:
            if drop_table:
                await self.client.drop(table.name())
            else:
                await self.remove_by(table.partial_init())
